@@ -19,14 +19,13 @@
  */
 
 #ifndef lint
-static char rcsid[] = "$Id: cbcp.c,v 1.2 1997/04/30 05:50:26 paulus Exp $";
+static char rcsid[] = "$Id: cbcp.c,v 1.6 1999/03/16 22:54:38 paulus Exp $";
 #endif
 
 #include <stdio.h>
 #include <string.h>
 #include <sys/types.h>
 #include <sys/time.h>
-#include <syslog.h>
 
 #include "pppd.h"
 #include "cbcp.h"
@@ -34,6 +33,17 @@ static char rcsid[] = "$Id: cbcp.c,v 1.2 1997/04/30 05:50:26 paulus Exp $";
 #include "lcp.h"
 #include "ipcp.h"
 
+/*
+ * Options.
+ */
+static int setcbcp __P((char **));
+
+static option_t cbcp_option_list[] = {
+    { "callback", o_special, setcbcp,
+      "Ask for callback" },
+    { NULL }
+};
+
 /*
  * Protocol entry points.
  */
@@ -60,6 +70,8 @@ struct protent cbcp_protent = {
     0,
     "CBCP",
     NULL,
+    cbcp_option_list,
+    NULL,
     NULL,
     NULL
 };
@@ -74,6 +86,21 @@ static void cbcp_up __P((cbcp_state *us));
 static void cbcp_recvack __P((cbcp_state *us, char *pckt, int len));
 static void cbcp_send __P((cbcp_state *us, u_char code, u_char *buf, int len));
 
+/* option processing */
+static int
+setcbcp(argv)
+    char **argv;
+{
+    lcp_wantoptions[0].neg_cbcp = 1;
+    cbcp_protent.enabled_flag = 1;
+    cbcp[0].us_number = strdup(*argv);
+    if (cbcp[0].us_number == 0)
+	novm("callback number");
+    cbcp[0].us_type |= (1 << CB_CONF_USER);
+    cbcp[0].us_type |= (1 << CB_CONF_ADMIN);
+    return (1);
+}
+
 /* init state */
 static void
 cbcp_init(iface)
@@ -94,18 +121,18 @@ cbcp_lowerup(iface)
 {
     cbcp_state *us = &cbcp[iface];
 
-    syslog(LOG_DEBUG, "cbcp_lowerup");
-    syslog(LOG_DEBUG, "want: %d", us->us_type);
+    dbglog("cbcp_lowerup");
+    dbglog("want: %d", us->us_type);
 
     if (us->us_type == CB_CONF_USER)
-        syslog(LOG_DEBUG, "phone no: %s", us->us_number);
+        dbglog("phone no: %s", us->us_number);
 }
 
 static void
 cbcp_open(unit)
     int unit;
 {
-    syslog(LOG_DEBUG, "cbcp_open");
+    dbglog("cbcp_open");
 }
 
 /* process an incomming packet */
@@ -124,7 +151,7 @@ cbcp_input(unit, inpacket, pktlen)
     inp = inpacket;
 
     if (pktlen < CBCP_MINLEN) {
-        syslog(LOG_ERR, "CBCP packet is too small");
+        error("CBCP packet is too small");
 	return;
     }
 
@@ -134,7 +161,7 @@ cbcp_input(unit, inpacket, pktlen)
 
 #if 0
     if (len > pktlen) {
-        syslog(LOG_ERR, "CBCP packet: invalid length");
+        error("CBCP packet: invalid length");
         return;
     }
 #endif
@@ -148,12 +175,12 @@ cbcp_input(unit, inpacket, pktlen)
 	break;
 
     case CBCP_RESP:
-	syslog(LOG_DEBUG, "CBCP_RESP received");
+	dbglog("CBCP_RESP received");
 	break;
 
     case CBCP_ACK:
 	if (id != us->us_id)
-	    syslog(LOG_DEBUG, "id doesn't match: expected %d recv %d",
+	    dbglog("id doesn't match: expected %d recv %d",
 		   us->us_id, id);
 
 	cbcp_recvack(us, inp, len);
@@ -272,7 +299,7 @@ cbcp_recvreq(us, pckt, pcktlen)
     address[0] = 0;
 
     while (len) {
-        syslog(LOG_DEBUG, "length: %d", len);
+        dbglog("length: %d", len);
 
 	GETCHAR(type, pckt);
 	GETCHAR(opt_len, pckt);
@@ -284,22 +311,22 @@ cbcp_recvreq(us, pckt, pcktlen)
 
 	switch(type) {
 	case CB_CONF_NO:
-	    syslog(LOG_DEBUG, "no callback allowed");
+	    dbglog("no callback allowed");
 	    break;
 
 	case CB_CONF_USER:
-	    syslog(LOG_DEBUG, "user callback allowed");
+	    dbglog("user callback allowed");
 	    if (opt_len > 4) {
 	        GETCHAR(addr_type, pckt);
 		memcpy(address, pckt, opt_len - 4);
 		address[opt_len - 4] = 0;
 		if (address[0])
-		    syslog(LOG_DEBUG, "address: %s", address);
+		    dbglog("address: %s", address);
 	    }
 	    break;
 
 	case CB_CONF_ADMIN:
-	    syslog(LOG_DEBUG, "user admin defined allowed");
+	    dbglog("user admin defined allowed");
 	    break;
 
 	case CB_CONF_LIST:
@@ -321,7 +348,7 @@ cbcp_resp(us)
     int len = 0;
 
     cb_type = us->us_allowed & us->us_type;
-    syslog(LOG_DEBUG, "cbcp_resp cb_type=%d", cb_type);
+    dbglog("cbcp_resp cb_type=%d", cb_type);
 
 #if 0
     if (!cb_type)
@@ -329,7 +356,7 @@ cbcp_resp(us)
 #endif
 
     if (cb_type & ( 1 << CB_CONF_USER ) ) {
-	syslog(LOG_DEBUG, "cbcp_resp CONF_USER");
+	dbglog("cbcp_resp CONF_USER");
 	PUTCHAR(CB_CONF_USER, bufp);
 	len = 3 + 1 + strlen(us->us_number) + 1;
 	PUTCHAR(len , bufp);
@@ -341,18 +368,17 @@ cbcp_resp(us)
     }
 
     if (cb_type & ( 1 << CB_CONF_ADMIN ) ) {
-	syslog(LOG_DEBUG, "cbcp_resp CONF_ADMIN");
+	dbglog("cbcp_resp CONF_ADMIN");
         PUTCHAR(CB_CONF_ADMIN, bufp);
-	len = 3 + 1;
-	PUTCHAR(len , bufp);
+	len = 3;
+	PUTCHAR(len, bufp);
 	PUTCHAR(5, bufp); /* delay */
-	PUTCHAR(0, bufp);
 	cbcp_send(us, CBCP_RESP, buf, len);
 	return;
     }
 
     if (cb_type & ( 1 << CB_CONF_NO ) ) {
-        syslog(LOG_DEBUG, "cbcp_resp CONF_NO");
+        dbglog("cbcp_resp CONF_NO");
 	PUTCHAR(CB_CONF_NO, bufp);
 	len = 3;
 	PUTCHAR(len , bufp);
@@ -411,15 +437,13 @@ cbcp_recvack(us, pckt, len)
 	    memcpy(address, pckt, opt_len - 4);
 	    address[opt_len - 4] = 0;
 	    if (address[0])
-	        syslog(LOG_DEBUG, "peer will call: %s", address);
+	        dbglog("peer will call: %s", address);
 	}
     }
 
     cbcp_up(us);
 }
 
-extern int persist;
-
 /* ok peer will do callback */
 static void
 cbcp_up(us)
