@@ -16,7 +16,7 @@
  * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
  * WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
  *
- * $Id: pppd.h,v 1.21 1998/03/26 04:46:08 paulus Exp $
+ * $Id: pppd.h,v 1.37 1999/03/31 05:39:43 paulus Exp $
  */
 
 /*
@@ -27,6 +27,7 @@
 #define __PPPD_H__
 
 #include <stdio.h>		/* for FILE */
+#include <limits.h>		/* for NGROUPS_MAX */
 #include <sys/param.h>		/* for MAXPATHLEN and BSD4_4, if defined */
 #include <sys/types.h>		/* for u_int32_t, if defined */
 #include <sys/time.h>		/* for struct timeval */
@@ -51,6 +52,75 @@
 #define MAXNAMELEN	256	/* max length of hostname or name for auth */
 #define MAXSECRETLEN	256	/* max length of password or secret */
 
+/*
+ * Option descriptor structure.
+ */
+
+typedef unsigned char	bool;
+
+enum opt_type {
+	o_special_noarg = 0,
+	o_special = 1,
+	o_bool,
+	o_int,
+	o_uint32,
+	o_string,
+};
+
+typedef struct {
+	char	*name;		/* name of the option */
+	enum opt_type type;
+	void	*addr;
+	char	*description;
+	int	flags;
+	void	*addr2;
+	int	upper_limit;
+	int	lower_limit;
+} option_t;
+
+/* Values for flags */
+#define OPT_VALUE	0xff	/* mask for presupplied value */
+#define OPT_HEX		0x100	/* int option is in hex */
+#define OPT_NOARG	0x200	/* option doesn't take argument */
+#define OPT_OR		0x400	/* OR in argument to value */
+#define OPT_INC		0x800	/* increment value */
+#define OPT_PRIV	0x1000	/* privileged option */
+#define OPT_STATIC	0x2000	/* string option goes into static array */
+#define OPT_LLIMIT	0x4000	/* check value against lower limit */
+#define OPT_ULIMIT	0x8000	/* check value against upper limit */
+#define OPT_LIMITS	(OPT_LLIMIT|OPT_ULIMIT)
+#define OPT_ZEROOK	0x10000	/* 0 value is OK even if not within limits */
+#define OPT_NOINCR	0x20000	/* value mustn't be increased */
+#define OPT_ZEROINF	0x40000	/* with OPT_NOINCR, 0 == infinity */
+#define OPT_A2INFO	0x100000 /* addr2 -> option_info to update */
+#define OPT_A2COPY	0x200000 /* addr2 -> second location to rcv value */
+#define OPT_ENABLE	0x400000 /* use *addr2 as enable for option */
+#define OPT_PRIVFIX	0x800000 /* can't be overridden if noauth */
+
+#define OPT_VAL(x)	((x) & OPT_VALUE)
+
+#ifndef GIDSET_TYPE
+#define GIDSET_TYPE	gid_t
+#endif
+
+/* Structure representing a list of permitted IP addresses. */
+struct permitted_ip {
+    int		permit;		/* 1 = permit, 0 = forbid */
+    u_int32_t	base;		/* match if (addr & mask) == base */
+    u_int32_t	mask;		/* base and mask are in network byte order */
+};
+
+/*
+ * Unfortunately, the linux kernel driver uses a different structure
+ * for statistics from the rest of the ports.
+ * This structure serves as a common representation for the bits
+ * pppd needs.
+ */
+struct pppd_stats {
+    unsigned int	bytes_in;
+    unsigned int	bytes_out;
+};
+
 /*
  * Global variables.
  */
@@ -70,6 +140,11 @@ extern int	privileged;	/* We were run by real-uid root */
 extern int	need_holdoff;	/* Need holdoff period after link terminates */
 extern char	**script_env;	/* Environment variables for scripts */
 extern int	detached;	/* Have detached from controlling tty */
+extern GIDSET_TYPE groups[NGROUPS_MAX];	/* groups the user is in */
+extern int	ngroups;	/* How many groups valid in groups */
+extern struct pppd_stats link_stats; /* byte/packet counts etc. for link */
+extern int	link_stats_valid; /* set if link_stats is valid */
+extern int	using_pty;	/* using pty as device (notty or pty opt.) */
 
 /*
  * Variables set by command-line options.
@@ -78,45 +153,47 @@ extern int	detached;	/* Have detached from controlling tty */
 extern int	debug;		/* Debug flag */
 extern int	kdebugflag;	/* Tell kernel to print debug messages */
 extern int	default_device;	/* Using /dev/tty or equivalent */
-extern char	devnam[];	/* Device name */
+extern char	devnam[MAXPATHLEN];	/* Device name */
 extern int	crtscts;	/* Use hardware flow control */
-extern int	modem;		/* Use modem control lines */
+extern bool	modem;		/* Use modem control lines */
 extern int	inspeed;	/* Input/Output speed requested */
 extern u_int32_t netmask;	/* IP netmask to set on interface */
-extern int	lockflag;	/* Create lock file to lock the serial dev */
-extern int	nodetach;	/* Don't detach from controlling tty */
+extern bool	lockflag;	/* Create lock file to lock the serial dev */
+extern bool	nodetach;	/* Don't detach from controlling tty */
+extern bool	updetach;	/* Detach from controlling tty when link up */
 extern char	*connector;	/* Script to establish physical link */
 extern char	*disconnector;	/* Script to disestablish physical link */
 extern char	*welcomer;	/* Script to welcome client after connection */
+extern char	*ptycommand;	/* Command to run on other side of pty */
 extern int	maxconnect;	/* Maximum connect time (seconds) */
-extern char	user[];		/* Our name for authenticating ourselves */
-extern char	passwd[];	/* Password for PAP */
-extern int	auth_required;	/* Peer is required to authenticate */
-extern int	proxyarp;	/* Set up proxy ARP entry for peer */
-extern int	persist;	/* Reopen link after it goes down */
-extern int	uselogin;	/* Use /etc/passwd for checking PAP */
-extern int	lcp_echo_interval; /* Interval between LCP echo-requests */
-extern int	lcp_echo_fails;	/* Tolerance to unanswered echo-requests */
-extern char	our_name[];	/* Our name for authentication purposes */
-extern char	remote_name[];	/* Peer's name for authentication */
-extern int	explicit_remote;/* remote_name specified with remotename opt */
-extern int	usehostname;	/* Use hostname for our_name */
-extern int	disable_defaultip; /* Don't use hostname for default IP adrs */
-extern int	demand;		/* Do dial-on-demand */
+extern char	user[MAXNAMELEN];/* Our name for authenticating ourselves */
+extern char	passwd[MAXSECRETLEN];	/* Password for PAP */
+extern bool	auth_required;	/* Peer is required to authenticate */
+extern bool	persist;	/* Reopen link after it goes down */
+extern bool	uselogin;	/* Use /etc/passwd for checking PAP */
+extern char	our_name[MAXNAMELEN];/* Our name for authentication purposes */
+extern char	remote_name[MAXNAMELEN]; /* Peer's name for authentication */
+extern bool	explicit_remote;/* remote_name specified with remotename opt */
+extern bool	demand;		/* Do dial-on-demand */
 extern char	*ipparam;	/* Extra parameter for ip up/down scripts */
-extern int	cryptpap;	/* Others' PAP passwords are encrypted */
+extern bool	cryptpap;	/* Others' PAP passwords are encrypted */
 extern int	idle_time_limit;/* Shut down link if idle for this long */
 extern int	holdoff;	/* Dead time before restarting */
-extern int	refuse_pap;	/* Don't wanna auth. ourselves with PAP */
-extern int	refuse_chap;	/* Don't wanna auth. ourselves with CHAP */
+extern bool	notty;		/* Stdin/out is not a tty */
+extern char	*record_file;	/* File to record chars sent/received */
+extern bool	sync_serial;	/* Device is synchronous serial device */
+
 #ifdef PPP_FILTER
 extern struct	bpf_program pass_filter;   /* Filter for pkts to pass */
 extern struct	bpf_program active_filter; /* Filter for link-active pkts */
 #endif
 
+char *current_option;		/* the name of the option being parsed */
+int  privileged_option;		/* set iff the current option came from root */
+char *option_source;		/* string saying where the option came from */
 
 #ifdef MSLANMAN
-extern int	ms_lanman;	/* Nonzero if use LanMan password instead of NT */
+extern bool	ms_lanman;	/* Use LanMan password instead of NT */
 				/* Has meaning only with MS-CHAP challenges */
 #endif
 
@@ -159,8 +236,10 @@ struct protent {
 			  void *arg));
     /* Process a received data packet */
     void (*datainput) __P((int unit, u_char *pkt, int len));
-    int  enabled_flag;		/* 0 iff protocol is disabled */
+    bool enabled_flag;		/* 0 iff protocol is disabled */
     char *name;			/* Text name of protocol */
+    char *data_name;		/* Text name of corresponding data protocol */
+    option_t *options;		/* List of command-line options */
     /* Check requested options, assign defaults */
     void (*check_options) __P((void));
     /* Configure interface for demand-dial */
@@ -185,20 +264,28 @@ void timeout __P((void (*func)(void *), void *arg, int t));
 				/* Call func(arg) after t seconds */
 void untimeout __P((void (*func)(void *), void *arg));
 				/* Cancel call to func(arg) */
-int run_program __P((char *prog, char **args, int must_exist));
+pid_t run_program __P((char *prog, char **args, int must_exist,
+		       void (*done)(void *), void *arg));
 				/* Run program prog with args in child */
 void demuxprotrej __P((int, int));
 				/* Demultiplex a Protocol-Reject */
-void format_packet __P((u_char *, int, void (*) (void *, char *, ...),
-		void *));	/* Format a packet in human-readable form */
 void log_packet __P((u_char *, int, char *, int));
 				/* Format a packet and log it with syslog */
 void print_string __P((char *, int,  void (*) (void *, char *, ...),
 		void *));	/* Format a string for output */
-int fmtmsg __P((char *, int, char *, ...));		/* sprintf++ */
-int vfmtmsg __P((char *, int, char *, va_list));	/* vsprintf++ */
+int slprintf __P((char *, int, char *, ...));		/* sprintf++ */
+int vslprintf __P((char *, int, char *, va_list));	/* vsprintf++ */
 void script_setenv __P((char *, char *));	/* set script env var */
 void script_unsetenv __P((char *));		/* unset script env var */
+size_t strlcpy __P((char *, const char *, size_t));	/* safe strcpy */
+size_t strlcat __P((char *, const char *, size_t));	/* safe strncpy */
+void dbglog __P((char *, ...));	/* log a debug message */
+void info __P((char *, ...));	/* log an informational message */
+void notice __P((char *, ...));	/* log a notice-level message */
+void warn __P((char *, ...));	/* log a warning message */
+void error __P((char *, ...));	/* log an error message */
+void fatal __P((char *, ...));	/* log an error message and die(1) */
+void reopen_log __P((void));	/* (re)open the connection to syslog */
 
 /* Procedures exported from auth.c */
 void link_required __P((int));	  /* we are starting to use the link */
@@ -227,8 +314,6 @@ int  auth_ip_addr __P((int, u_int32_t));
 				/* check if IP address is authorized */
 int  bad_ip_adrs __P((u_int32_t));
 				/* check if IP address is unreasonable */
-void check_access __P((FILE *, char *));
-				/* check permissions on secrets file */
 
 /* Procedures exported from demand.c */
 void demand_conf __P((void));	/* config interface(s) for demand-dial */
@@ -237,16 +322,17 @@ void demand_unblock __P((void)); /* set all NPs to pass packets */
 void demand_discard __P((void)); /* set all NPs to discard packets */
 void demand_rexmit __P((int));	/* retransmit saved frames for an NP */
 int  loop_chars __P((unsigned char *, int)); /* process chars from loopback */
-int  loop_frame __P((unsigned char *, int)); /* process frame from loopback */
+int  loop_frame __P((unsigned char *, int)); /* should we bring link up? */
 
 /* Procedures exported from sys-*.c */
 void sys_init __P((void));	/* Do system-dependent initialization */
 void sys_cleanup __P((void));	/* Restore system state before exiting */
-void sys_check_options __P((void)); /* Check options specified */
+int  sys_check_options __P((void)); /* Check options specified */
 void sys_close __P((void));	/* Clean up in a child before execing */
 int  ppp_available __P((void));	/* Test whether ppp kernel support exists */
-void open_ppp_loopback __P((void)); /* Open loopback for demand-dialling */
-void establish_ppp __P((int));	/* Turn serial port into a ppp interface */
+int  get_pty __P((int *, int *, char *, int));	/* Get pty master/slave */
+int  open_ppp_loopback __P((void)); /* Open loopback for demand-dialling */
+int  establish_ppp __P((int));	/* Turn serial port into a ppp interface */
 void restore_loop __P((void));	/* Transfer ppp unit back to loopback */
 void disestablish_ppp __P((int)); /* Restore port to normal operation */
 void clean_check __P((void));	/* Check if line was 8-bit clean */
@@ -256,9 +342,8 @@ void setdtr __P((int, int));	/* Raise or lower port's DTR line */
 void output __P((int, u_char *, int)); /* Output a PPP packet */
 void wait_input __P((struct timeval *));
 				/* Wait for input, with timeout */
-void wait_loop_output __P((struct timeval *));
-				/* Wait for pkt from loopback, with timeout */
-void wait_time __P((struct timeval *)); /* Wait for given length of time */
+void add_fd __P((int));		/* Add fd to set to wait for */
+void remove_fd __P((int));	/* Remove fd from set to wait for */
 int  read_packet __P((u_char *)); /* Read PPP packet */
 int  get_loop_output __P((void)); /* Read pkts from loopback */
 void ppp_send_config __P((int, int, u_int32_t, int, int));
@@ -274,6 +359,8 @@ void ccp_flags_set __P((int, int, int));
 int  ccp_fatal_error __P((int)); /* Test for fatal decomp error in kernel */
 int  get_idle_time __P((int, struct ppp_idle *));
 				/* Find out how long link has been idle */
+int  get_ppp_stats __P((int, struct pppd_stats *));
+				/* Return link statistics */
 int  sifvjcomp __P((int, int, int, int));
 				/* Configure VJ TCP header compression */
 int  sifup __P((int));		/* Configure i/f up (for IP) */
@@ -299,15 +386,19 @@ int  daemon __P((int, int));	/* Detach us from terminal session */
 void logwtmp __P((const char *, const char *, const char *));
 				/* Write entry to wtmp file */
 int  get_host_seed __P((void));	/* Get host-dependent random number seed */
+int  have_route_to __P((u_int32_t)); /* Check if route to addr exists */
 #ifdef PPP_FILTER
 int  set_filters __P((struct bpf_program *pass, struct bpf_program *active));
 				/* Set filter programs in kernel */
 #endif
+#ifdef IPX_CHANGE
+int  sipxfaddr __P((int, unsigned long, unsigned char *));
+int  cipxfaddr __P((int));
+#endif
 
 /* Procedures exported from options.c */
 int  parse_args __P((int argc, char **argv));
 				/* Parse options from arguments given */
-void usage __P((void));		/* Print a usage message */
 int  options_from_file __P((char *filename, int must_exist, int check_prot,
 			    int privileged));
 				/* Parse options from an options file */
@@ -319,6 +410,8 @@ int  getword __P((FILE *f, char *word, int *newlinep, char *filename));
 				/* Read a word from a file */
 void option_error __P((char *fmt, ...));
 				/* Print an error message about an option */
+int int_option __P((char *, int *));
+				/* Simplified number_option for decimal ints */
 
 /*
  * This structure is used to store information about certain
@@ -331,11 +424,11 @@ struct option_info {
     char    *source;		/* where option came from */
 };
 
-extern struct option_info auth_req_info;
+extern struct option_info devnam_info;
 extern struct option_info connector_info;
 extern struct option_info disconnector_info;
 extern struct option_info welcomer_info;
-extern struct option_info devnam_info;
+extern struct option_info ptycommand_info;
 
 /*
  * Inline versions of get/put char/short/long.
@@ -376,25 +469,17 @@ extern struct option_info devnam_info;
 #define INCPTR(n, cp)	((cp) += (n))
 #define DECPTR(n, cp)	((cp) -= (n))
 
-#undef  FALSE
-#define FALSE	0
-#undef  TRUE
-#define TRUE	1
-
 /*
  * System dependent definitions for user-level 4.3BSD UNIX implementation.
  */
 
-#define DEMUXPROTREJ(u, p)	demuxprotrej(u, p)
-
 #define TIMEOUT(r, f, t)	timeout((r), (f), (t))
 #define UNTIMEOUT(r, f)		untimeout((r), (f))
 
 #define BCOPY(s, d, l)		memcpy(d, s, l)
 #define BZERO(s, n)		memset(s, 0, n)
-#define EXIT(u)			quit()
 
-#define PRINTMSG(m, l)	{ m[l] = '\0'; syslog(LOG_INFO, "Remote message: %s", m); }
+#define PRINTMSG(m, l)		{ info("Remote message: %0.*v", l, m); }
 
 /*
  * MAKEHEADER - Add Header fields to a packet.
@@ -425,49 +510,49 @@ extern struct option_info devnam_info;
 #endif /* LOG_PPP */
 
 #ifdef DEBUGMAIN
-#define MAINDEBUG(x)	if (debug) syslog x
+#define MAINDEBUG(x)	if (debug) dbglog x
 #else
 #define MAINDEBUG(x)
 #endif
 
 #ifdef DEBUGSYS
-#define SYSDEBUG(x)	if (debug) syslog x
+#define SYSDEBUG(x)	if (debug) dbglog x
 #else
 #define SYSDEBUG(x)
 #endif
 
 #ifdef DEBUGFSM
-#define FSMDEBUG(x)	if (debug) syslog x
+#define FSMDEBUG(x)	if (debug) dbglog x
 #else
 #define FSMDEBUG(x)
 #endif
 
 #ifdef DEBUGLCP
-#define LCPDEBUG(x)	if (debug) syslog x
+#define LCPDEBUG(x)	if (debug) dbglog x
 #else
 #define LCPDEBUG(x)
 #endif
 
 #ifdef DEBUGIPCP
-#define IPCPDEBUG(x)	if (debug) syslog x
+#define IPCPDEBUG(x)	if (debug) dbglog x
 #else
 #define IPCPDEBUG(x)
 #endif
 
 #ifdef DEBUGUPAP
-#define UPAPDEBUG(x)	if (debug) syslog x
+#define UPAPDEBUG(x)	if (debug) dbglog x
 #else
 #define UPAPDEBUG(x)
 #endif
 
 #ifdef DEBUGCHAP
-#define CHAPDEBUG(x)	if (debug) syslog x
+#define CHAPDEBUG(x)	if (debug) dbglog x
 #else
 #define CHAPDEBUG(x)
 #endif
 
 #ifdef DEBUGIPXCP
-#define IPXCPDEBUG(x)	if (debug) syslog x
+#define IPXCPDEBUG(x)	if (debug) dbglog x
 #else
 #define IPXCPDEBUG(x)
 #endif
