@@ -21,7 +21,7 @@
  */
 
 #ifndef lint
-static char rcsid[] = "$Id: sys-bsd.c,v 1.31 1998/04/02 12:04:19 paulus Exp $";
+static char rcsid[] = "$Id: sys-bsd.c,v 1.42 1999/04/01 07:20:10 paulus Exp $";
 /*	$NetBSD: sys-bsd.c,v 1.1.1.3 1997/09/26 18:53:04 christos Exp $	*/
 #endif
 
@@ -30,7 +30,6 @@ static char rcsid[] = "$Id: sys-bsd.c,v 1.31 1998/04/02 12:04:19 paulus Exp $";
  */
 
 #include <stdio.h>
-#include <syslog.h>
 #include <string.h>
 #include <stdlib.h>
 #include <unistd.h>
@@ -94,6 +93,9 @@ static unsigned char inbuf[512]; /* buffer for chars read from loopback */
 
 static int sockfd;		/* socket for doing interface ioctls */
 
+static fd_set in_fds;		/* set of fds that wait_input waits for */
+static int max_in_fd;		/* highest fd set in in_fds */
+
 static int if_is_up;		/* the interface is currently up */
 static u_int32_t ifaddrs[2];	/* local and remote addresses we set */
 static u_int32_t default_route_gateway;	/* gateway addr for default route */
@@ -111,10 +113,11 @@ void
 sys_init()
 {
     /* Get an internet socket for doing socket ioctl's on. */
-    if ((sockfd = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
-	syslog(LOG_ERR, "Couldn't create IP socket: %m");
-	die(1);
-    }
+    if ((sockfd = socket(AF_INET, SOCK_DGRAM, 0)) < 0)
+	fatal("Couldn't create IP socket: %m");
+
+    FD_ZERO(&in_fds);
+    max_in_fd = 0;
 }
 
 /*
@@ -128,7 +131,7 @@ sys_cleanup()
     struct ifreq ifr;
 
     if (if_is_up) {
-	strncpy(ifr.ifr_name, ifname, sizeof(ifr.ifr_name));
+	strlcpy(ifr.ifr_name, ifname, sizeof(ifr.ifr_name));
 	if (ioctl(sockfd, SIOCGIFFLAGS, &ifr) >= 0
 	    && ((ifr.ifr_flags & IFF_UP) != 0)) {
 	    ifr.ifr_flags &= ~IFF_UP;
@@ -159,9 +162,16 @@ sys_close()
 /*
  * sys_check_options - check the options that the user specified
  */
-void
+int
 sys_check_options()
 {
+#ifndef CDTRCTS
+    if (crtscts == 2) {
+	warn("DTR/CTS flow control is not supported on this system");
+	return 0;
+    }
+#endif
+    return 1;
 }
 
 /*
@@ -178,7 +188,7 @@ ppp_available()
     if ((s = socket(AF_INET, SOCK_DGRAM, 0)) < 0)
 	return 1;		/* can't tell */
 
-    strncpy(ifr.ifr_name, "ppp0", sizeof (ifr.ifr_name));
+    strlcpy(ifr.ifr_name, "ppp0", sizeof (ifr.ifr_name));
     ok = ioctl(s, SIOCGIFFLAGS, (caddr_t) &ifr) >= 0;
     close(s);
 
@@ -192,7 +202,7 @@ file in the ppp-2.2 distribution.\n";
 /*
  * establish_ppp - Turn the serial port into a ppp interface.
  */
-void
+int
 establish_ppp(fd)
     int fd;
 {
@@ -203,45 +213,32 @@ establish_ppp(fd)
 	/*
 	 * Demand mode - prime the old ppp device to relinquish the unit.
 	 */
-	if (ioctl(ppp_fd, PPPIOCXFERUNIT, 0) < 0) {
-	    syslog(LOG_ERR, "ioctl(transfer ppp unit): %m");
-	    die(1);
-	}
+	if (ioctl(ppp_fd, PPPIOCXFERUNIT, 0) < 0)
+	    fatal("ioctl(transfer ppp unit): %m");
     }
 
     /*
      * Save the old line discipline of fd, and set it to PPP.
      */
-    if (ioctl(fd, TIOCGETD, &initdisc) < 0) {
-	syslog(LOG_ERR, "ioctl(TIOCGETD): %m");
-	die(1);
-    }
-    if (ioctl(fd, TIOCSETD, &pppdisc) < 0) {
-	syslog(LOG_ERR, "ioctl(TIOCSETD): %m");
-	die(1);
-    }
+    if (ioctl(fd, TIOCGETD, &initdisc) < 0)
+	fatal("ioctl(TIOCGETD): %m");
+    if (ioctl(fd, TIOCSETD, &pppdisc) < 0)
+	fatal("ioctl(TIOCSETD): %m");
 
     if (!demand) {
 	/*
 	 * Find out which interface we were given.
 	 */
-	if (ioctl(fd, PPPIOCGUNIT, &ifunit) < 0) {	
-	    syslog(LOG_ERR, "ioctl(PPPIOCGUNIT): %m");
-	    die(1);
-	}
+	if (ioctl(fd, PPPIOCGUNIT, &ifunit) < 0) {    
+	    fatal("ioctl(PPPIOCGUNIT): %m");
     } else {
 	/*
 	 * Check that we got the same unit again.
 	 */
-	if (ioctl(fd, PPPIOCGUNIT, &x) < 0) {	
-	    syslog(LOG_ERR, "ioctl(PPPIOCGUNIT): %m");
-	    die(1);
-	}
-	if (x != ifunit) {
-	    syslog(LOG_ERR, "transfer_ppp failed: wanted unit %d, got %d",
-		   ifunit, x);
-	    die(1);
-	}
+	if (ioctl(fd, PPPIOCGUNIT, &x) < 0) { 
+	    fatal("ioctl(PPPIOCGUNIT): %m");
+	if (x != ifunit)
+	    fatal("transfer_ppp failed: wanted unit %d, got %d", ifunit, x);
 	x = TTYDISC;
 	ioctl(loop_slave, TIOCSETD, &x);
     }
@@ -253,11 +250,11 @@ establish_ppp(fd)
      */
     if (kdebugflag) {
 	if (ioctl(fd, PPPIOCGFLAGS, (caddr_t) &x) < 0) {
-	    syslog(LOG_WARNING, "ioctl (PPPIOCGFLAGS): %m");
+	    warn("ioctl (PPPIOCGFLAGS): %m");
 	} else {
 	    x |= (kdebugflag & 0xFF) * SC_DEBUG;
 	    if (ioctl(fd, PPPIOCSFLAGS, (caddr_t) &x) < 0)
-		syslog(LOG_WARNING, "ioctl(PPPIOCSFLAGS): %m");
+		warn("ioctl(PPPIOCSFLAGS): %m");
 	}
     }
 
@@ -266,8 +263,10 @@ establish_ppp(fd)
      */
     if ((initfdflags = fcntl(fd, F_GETFL)) == -1
 	|| fcntl(fd, F_SETFL, initfdflags | O_NONBLOCK) == -1) {
-	syslog(LOG_WARNING, "Couldn't set device to non-blocking mode: %m");
+	warn("Couldn't set device to non-blocking mode: %m");
     }
+
+    return fd;
 }
 
 /*
@@ -281,28 +280,19 @@ restore_loop()
     /*
      * Transfer the ppp interface back to the loopback.
      */
-    if (ioctl(ppp_fd, PPPIOCXFERUNIT, 0) < 0) {
-	syslog(LOG_ERR, "ioctl(transfer ppp unit): %m");
-	die(1);
-    }
+    if (ioctl(ppp_fd, PPPIOCXFERUNIT, 0) < 0)
+	fatal("ioctl(transfer ppp unit): %m");
     x = PPPDISC;
-    if (ioctl(loop_slave, TIOCSETD, &x) < 0) {
-	syslog(LOG_ERR, "ioctl(TIOCSETD): %m");
-	die(1);
-    }
+    if (ioctl(loop_slave, TIOCSETD, &x) < 0)
+	fatal("ioctl(TIOCSETD): %m");
 
     /*
      * Check that we got the same unit again.
      */
-    if (ioctl(loop_slave, PPPIOCGUNIT, &x) < 0) {	
-	syslog(LOG_ERR, "ioctl(PPPIOCGUNIT): %m");
-	die(1);
-    }
-    if (x != ifunit) {
-	syslog(LOG_ERR, "transfer_ppp failed: wanted unit %d, got %d",
-	       ifunit, x);
-	die(1);
-    }
+    if (ioctl(loop_slave, PPPIOCGUNIT, &x) < 0)
+	fatal("ioctl(PPPIOCGUNIT): %m");
+    if (x != ifunit)
+	fatal("transfer_ppp failed: wanted unit %d, got %d", ifunit, x);
     ppp_fd = loop_slave;
 }
 
@@ -317,12 +307,12 @@ disestablish_ppp(fd)
 {
     /* Reset non-blocking mode on fd. */
     if (initfdflags != -1 && fcntl(fd, F_SETFL, initfdflags) < 0)
-	syslog(LOG_WARNING, "Couldn't restore device fd flags: %m");
+	warn("Couldn't restore device fd flags: %m");
     initfdflags = -1;
 
     /* Restore old line discipline. */
     if (initdisc >= 0 && ioctl(fd, TIOCSETD, &initdisc) < 0)
-	syslog(LOG_ERR, "ioctl(TIOCSETD): %m");
+	error("ioctl(TIOCSETD): %m");
     initdisc = -1;
 
     if (fd == ppp_fd)
@@ -355,8 +345,8 @@ clean_check()
 	    break;
 	}
 	if (s != NULL) {
-	    syslog(LOG_WARNING, "Serial link is not 8-bit clean:");
-	    syslog(LOG_WARNING, "All received characters had %s", s);
+	    warn("Serial link is not 8-bit clean:");
+	    warn("All received characters had %s", s);
 	}
     }
 }
@@ -374,10 +364,8 @@ set_up_tty(fd, local)
 {
     struct termios tios;
 
-    if (tcgetattr(fd, &tios) < 0) {
-	syslog(LOG_ERR, "tcgetattr: %m");
-	die(1);
-    }
+    if (tcgetattr(fd, &tios) < 0)
+	fatal("tcgetattr: %m");
 
     if (!restore_term) {
 	inittermios = tios;
@@ -385,10 +373,19 @@ set_up_tty(fd, local)
     }
 
     tios.c_cflag &= ~(CSIZE | CSTOPB | PARENB | CLOCAL);
-    if (crtscts > 0 && !local)
-	tios.c_cflag |= CRTSCTS;
-    else if (crtscts < 0)
+    if (crtscts > 0 && !local) {
+        if (crtscts == 2) {
+#ifdef CDTRCTS
+            tios.c_cflag |= CDTRCTS;
+#endif
+	} else
+	    tios.c_cflag |= CRTSCTS;
+    } else if (crtscts < 0) {
 	tios.c_cflag &= ~CRTSCTS;
+#ifdef CDTRCTS
+	tios.c_cflag &= ~CDTRCTS;
+#endif
+    }
 
     tios.c_cflag |= CS8 | CREAD | HUPCL;
     if (local || !modem)
@@ -414,18 +411,13 @@ set_up_tty(fd, local)
 	 * We can't proceed if the serial port speed is 0,
 	 * since that implies that the serial port is disabled.
 	 */
-	if (inspeed == 0) {
-	    syslog(LOG_ERR, "Baud rate for %s is 0; need explicit baud rate",
-		   devnam);
-	    die(1);
-	}
+	if (inspeed == 0)
+	    fatal("Baud rate for %s is 0; need explicit baud rate", devnam);
     }
     baud_rate = inspeed;
 
-    if (tcsetattr(fd, TCSAFLUSH, &tios) < 0) {
-	syslog(LOG_ERR, "tcsetattr: %m");
-	die(1);
-    }
+    if (tcsetattr(fd, TCSAFLUSH, &tios) < 0)
+	fatal("tcsetattr: %m");
 
     restore_term = 1;
 }
@@ -449,7 +441,7 @@ restore_tty(fd)
 	}
 	if (tcsetattr(fd, TCSAFLUSH, &inittermios) < 0)
 	    if (errno != ENXIO)
-		syslog(LOG_WARNING, "tcsetattr: %m");
+		warn("tcsetattr: %m");
 	ioctl(fd, TIOCSWINSZ, &wsinfo);
 	restore_term = 0;
     }
@@ -468,24 +460,54 @@ int fd, on;
     ioctl(fd, (on? TIOCMBIS: TIOCMBIC), &modembits);
 }
 
+/*
+ * get_pty - get a pty master/slave pair and chown the slave side
+ * to the uid given.  Assumes slave_name points to >= 12 bytes of space.
+ */
+int
+get_pty(master_fdp, slave_fdp, slave_name, uid)
+    int *master_fdp;
+    int *slave_fdp;
+    char *slave_name;
+    int uid;
+{
+    struct termios tios;
+
+    if (openpty(master_fdp, slave_fdp, slave_name, NULL, NULL) < 0)
+	return 0;
+
+    fchown(*slave_fdp, uid, -1);
+    fchmod(*slave_fdp, S_IRUSR | S_IWUSR);
+    if (tcgetattr(*slave_fdp, &tios) == 0) {
+	tios.c_cflag &= ~(CSIZE | CSTOPB | PARENB);
+	tios.c_cflag |= CS8 | CREAD;
+	tios.c_iflag  = IGNPAR | CLOCAL;
+	tios.c_oflag  = 0;
+	tios.c_lflag  = 0;
+	if (tcsetattr(*slave_fdp, TCSAFLUSH, &tios) < 0)
+	    warn("couldn't set attributes on pty: %m");
+    } else
+	warn("couldn't get attributes on pty: %m");
+
+    return 1;
+}
+
 
 /*
  * open_ppp_loopback - open the device we use for getting
  * packets in demand mode, and connect it to a ppp interface.
  * Here we use a pty.
  */
-void
+int
 open_ppp_loopback()
 {
     int flags;
     struct termios tios;
     int pppdisc = PPPDISC;
 
-    if (openpty(&loop_master, &loop_slave, loop_name, NULL, NULL) < 0) {
-	syslog(LOG_ERR, "No free pty for loopback");
-	die(1);
-    }
-    SYSDEBUG((LOG_DEBUG, "using %s for loopback", loop_name));
+    if (openpty(&loop_master, &loop_slave, loop_name, NULL, NULL) < 0)
+	fatal("No free pty for loopback");
+    SYSDEBUG(("using %s for loopback", loop_name));
 
     if (tcgetattr(loop_slave, &tios) == 0) {
 	tios.c_cflag &= ~(CSIZE | CSTOPB | PARENB);
@@ -494,40 +516,37 @@ open_ppp_loopback()
 	tios.c_oflag = 0;
 	tios.c_lflag = 0;
 	if (tcsetattr(loop_slave, TCSAFLUSH, &tios) < 0)
-	    syslog(LOG_WARNING, "couldn't set attributes on loopback: %m");
+	    warn("couldn't set attributes on loopback: %m");
     }
 
     if ((flags = fcntl(loop_master, F_GETFL)) != -1) 
 	if (fcntl(loop_master, F_SETFL, flags | O_NONBLOCK) == -1)
-	    syslog(LOG_WARNING, "couldn't set loopback to nonblock: %m");
+	    warn("couldn't set loopback to nonblock: %m");
 
     ppp_fd = loop_slave;
-    if (ioctl(ppp_fd, TIOCSETD, &pppdisc) < 0) {
-	syslog(LOG_ERR, "ioctl(TIOCSETD): %m");
-	die(1);
-    }
+    if (ioctl(ppp_fd, TIOCSETD, &pppdisc) < 0)
+	fatal("ioctl(TIOCSETD): %m");
 
     /*
      * Find out which interface we were given.
      */
-    if (ioctl(ppp_fd, PPPIOCGUNIT, &ifunit) < 0) {	
-	syslog(LOG_ERR, "ioctl(PPPIOCGUNIT): %m");
-	die(1);
-    }
+    if (ioctl(ppp_fd, PPPIOCGUNIT, &ifunit) < 0)
+	fatal("ioctl(PPPIOCGUNIT): %m");
 
     /*
      * Enable debug in the driver if requested.
      */
     if (kdebugflag) {
 	if (ioctl(ppp_fd, PPPIOCGFLAGS, (caddr_t) &flags) < 0) {
-	    syslog(LOG_WARNING, "ioctl (PPPIOCGFLAGS): %m");
+	    warn("ioctl (PPPIOCGFLAGS): %m");
 	} else {
 	    flags |= (kdebugflag & 0xFF) * SC_DEBUG;
 	    if (ioctl(ppp_fd, PPPIOCSFLAGS, (caddr_t) &flags) < 0)
-		syslog(LOG_WARNING, "ioctl(PPPIOCSFLAGS): %m");
+		warn("ioctl(PPPIOCSFLAGS): %m");
 	}
     }
 
+    return loop_master;
 }
 
 
@@ -541,17 +560,17 @@ output(unit, p, len)
     int len;
 {
     if (debug)
-	log_packet(p, len, "sent ", LOG_DEBUG);
+	dbglog("sent %P", p, len);
 
     if (write(ttyfd, p, len) < 0) {
 	if (errno != EIO)
-	    syslog(LOG_ERR, "write: %m");
+	    error("write: %m");
     }
 }
 
 
 /*
- * wait_input - wait until there is data available on ttyfd,
+ * wait_input - wait until there is data available,
  * for the length of time specified by *timo (indefinite
  * if timo is NULL).
  */
@@ -562,16 +581,34 @@ wait_input(timo)
     fd_set ready;
     int n;
 
-    FD_ZERO(&ready);
-    FD_SET(ttyfd, &ready);
-    n = select(ttyfd+1, &ready, NULL, &ready, timo);
-    if (n < 0 && errno != EINTR) {
-	syslog(LOG_ERR, "select: %m");
-	die(1);
-    }
+    ready = in_fds;
+    n = select(max_in_fd + 1, &ready, NULL, &ready, timo);
+    if (n < 0 && errno != EINTR)
+	fatal("select: %m");
 }
 
 
+/*
+ * add_fd - add an fd to the set that wait_input waits for.
+ */
+void add_fd(fd)
+    int fd;
+{
+    FD_SET(fd, &in_fds);
+    if (fd > max_in_fd)
+	max_in_fd = fd;
+}
+
+/*
+ * remove_fd - remove an fd from the set that wait_input waits for.
+ */
+void remove_fd(fd)
+    int fd;
+{
+    FD_CLR(fd, &in_fds);
+}
+
+#if 0
 /*
  * wait_loop_output - wait until there is data available on the
  * loopback, for the length of time specified by *timo (indefinite
@@ -587,10 +624,8 @@ wait_loop_output(timo)
     FD_ZERO(&ready);
     FD_SET(loop_master, &ready);
     n = select(loop_master + 1, &ready, NULL, &ready, timo);
-    if (n < 0 && errno != EINTR) {
-	syslog(LOG_ERR, "select: %m");
-	die(1);
-    }
+    if (n < 0 && errno != EINTR)
+	fatal("select: %m");
 }
 
 
@@ -605,11 +640,10 @@ wait_time(timo)
     int n;
 
     n = select(0, NULL, NULL, NULL, timo);
-    if (n < 0 && errno != EINTR) {
-	syslog(LOG_ERR, "select: %m");
-	die(1);
-    }
+    if (n < 0 && errno != EINTR)
+	fatal("select: %m");
 }
+#endif
 
 
 /*
@@ -624,8 +658,7 @@ read_packet(buf)
     if ((len = read(ttyfd, buf, PPP_MTU + PPP_HDRLEN)) < 0) {
 	if (errno == EWOULDBLOCK || errno == EINTR)
 	    return -1;
-	syslog(LOG_ERR, "read: %m");
-	die(1);
+	fatal("read: %m");
     }
     return len;
 }
@@ -647,13 +680,10 @@ get_loop_output()
 	    rv = 1;
     }
 
-    if (n == 0) {
-	syslog(LOG_ERR, "eof on loopback");
-	die(1);
-    } else if (errno != EWOULDBLOCK){
-	syslog(LOG_ERR, "read from loopback: %m");
-	die(1);
-    }
+    if (n == 0)
+	fatal("eof on loopback");
+    if (errno != EWOULDBLOCK)
+	fatal("read from loopback: %m");
 
     return rv;
 }
@@ -672,28 +702,20 @@ ppp_send_config(unit, mtu, asyncmap, pcomp, accomp)
     u_int x;
     struct ifreq ifr;
 
-    strncpy(ifr.ifr_name, ifname, sizeof (ifr.ifr_name));
+    strlcpy(ifr.ifr_name, ifname, sizeof (ifr.ifr_name));
     ifr.ifr_mtu = mtu;
-    if (ioctl(sockfd, SIOCSIFMTU, (caddr_t) &ifr) < 0) {
-	syslog(LOG_ERR, "ioctl(SIOCSIFMTU): %m");
-	quit();
-    }
+    if (ioctl(sockfd, SIOCSIFMTU, (caddr_t) &ifr) < 0)
+	fatal("ioctl(SIOCSIFMTU): %m");
 
-    if (ioctl(ppp_fd, PPPIOCSASYNCMAP, (caddr_t) &asyncmap) < 0) {
-	syslog(LOG_ERR, "ioctl(PPPIOCSASYNCMAP): %m");
-	quit();
-    }
+    if (ioctl(ppp_fd, PPPIOCSASYNCMAP, (caddr_t) &asyncmap) < 0)
+	fatal("ioctl(PPPIOCSASYNCMAP): %m");
 
-    if (ioctl(ppp_fd, PPPIOCGFLAGS, (caddr_t) &x) < 0) {
-	syslog(LOG_ERR, "ioctl (PPPIOCGFLAGS): %m");
-	quit();
-    }
+    if (ioctl(ppp_fd, PPPIOCGFLAGS, (caddr_t) &x) < 0)
+	fatal("ioctl (PPPIOCGFLAGS): %m");
     x = pcomp? x | SC_COMP_PROT: x &~ SC_COMP_PROT;
     x = accomp? x | SC_COMP_AC: x &~ SC_COMP_AC;
-    if (ioctl(ppp_fd, PPPIOCSFLAGS, (caddr_t) &x) < 0) {
-	syslog(LOG_ERR, "ioctl(PPPIOCSFLAGS): %m");
-	quit();
-    }
+    if (ioctl(ppp_fd, PPPIOCSFLAGS, (caddr_t) &x) < 0)
+	fatal("ioctl(PPPIOCSFLAGS): %m");
 }
 
 
@@ -706,7 +728,7 @@ ppp_set_xaccm(unit, accm)
     ext_accm accm;
 {
     if (ioctl(ppp_fd, PPPIOCSXASYNCMAP, accm) < 0 && errno != ENOTTY)
-	syslog(LOG_WARNING, "ioctl(set extended ACCM): %m");
+	warn("ioctl(set extended ACCM): %m");
 }
 
 
@@ -722,23 +744,15 @@ ppp_recv_config(unit, mru, asyncmap, pcomp, accomp)
 {
     int x;
 
-    if (ioctl(ppp_fd, PPPIOCSMRU, (caddr_t) &mru) < 0) {
-	syslog(LOG_ERR, "ioctl(PPPIOCSMRU): %m");
-	quit();
-    }
-    if (ioctl(ppp_fd, PPPIOCSRASYNCMAP, (caddr_t) &asyncmap) < 0) {
-	syslog(LOG_ERR, "ioctl(PPPIOCSRASYNCMAP): %m");
-	quit();
-    }
-    if (ioctl(ppp_fd, PPPIOCGFLAGS, (caddr_t) &x) < 0) {
-	syslog(LOG_ERR, "ioctl (PPPIOCGFLAGS): %m");
-	quit();
-    }
+    if (ioctl(ppp_fd, PPPIOCSMRU, (caddr_t) &mru) < 0)
+	fatal("ioctl(PPPIOCSMRU): %m");
+    if (ioctl(ppp_fd, PPPIOCSRASYNCMAP, (caddr_t) &asyncmap) < 0)
+	fatal("ioctl(PPPIOCSRASYNCMAP): %m");
+    if (ioctl(ppp_fd, PPPIOCGFLAGS, (caddr_t) &x) < 0)
+	fatal("ioctl (PPPIOCGFLAGS): %m");
     x = !accomp? x | SC_REJ_COMP_AC: x &~ SC_REJ_COMP_AC;
-    if (ioctl(ppp_fd, PPPIOCSFLAGS, (caddr_t) &x) < 0) {
-	syslog(LOG_ERR, "ioctl(PPPIOCSFLAGS): %m");
-	quit();
-    }
+    if (ioctl(ppp_fd, PPPIOCSFLAGS, (caddr_t) &x) < 0)
+	fatal("ioctl(PPPIOCSFLAGS): %m");
 }
 
 /*
@@ -772,13 +786,13 @@ ccp_flags_set(unit, isopen, isup)
     int x;
 
     if (ioctl(ppp_fd, PPPIOCGFLAGS, (caddr_t) &x) < 0) {
-	syslog(LOG_ERR, "ioctl (PPPIOCGFLAGS): %m");
+	error("ioctl (PPPIOCGFLAGS): %m");
 	return;
     }
     x = isopen? x | SC_CCP_OPEN: x &~ SC_CCP_OPEN;
     x = isup? x | SC_CCP_UP: x &~ SC_CCP_UP;
     if (ioctl(ppp_fd, PPPIOCSFLAGS, (caddr_t) &x) < 0)
-	syslog(LOG_ERR, "ioctl(PPPIOCSFLAGS): %m");
+	error("ioctl(PPPIOCSFLAGS): %m");
 }
 
 /*
@@ -793,7 +807,7 @@ ccp_fatal_error(unit)
     int x;
 
     if (ioctl(ppp_fd, PPPIOCGFLAGS, (caddr_t) &x) < 0) {
-	syslog(LOG_ERR, "ioctl(PPPIOCGFLAGS): %m");
+	error("ioctl(PPPIOCGFLAGS): %m");
 	return 0;
     }
     return x & SC_DC_FERROR;
@@ -810,6 +824,27 @@ get_idle_time(u, ip)
     return ioctl(ppp_fd, PPPIOCGIDLE, ip) >= 0;
 }
 
+/*
+ * get_ppp_stats - return statistics for the link.
+ */
+int
+get_ppp_stats(u, stats)
+    int u;
+    struct pppd_stats *stats;
+{
+    struct ifpppstatsreq req;
+
+    memset (&req, 0, sizeof (req));
+    strlcpy(req.ifr_name, interface, sizeof(req.ifr_name));
+    if (ioctl(sockfd, SIOCGPPPSTATS, &req) < 0) {
+	error("Couldn't get PPP statistics: %m");
+	return 0;
+    }
+    stats->bytes_in = req.stats.p.ppp_ibytes;
+    stats->bytes_out = req.stats.p.ppp_obytes;
+    return 1;
+}
+
 
 #ifdef PPP_FILTER
 /*
@@ -823,13 +858,13 @@ set_filters(pass, active)
 
     if (pass->bf_len > 0) {
 	if (ioctl(ppp_fd, PPPIOCSPASS, pass) < 0) {
-	    syslog(LOG_ERR, "Couldn't set pass-filter in kernel: %m");
+	    error("Couldn't set pass-filter in kernel: %m");
 	    ret = 0;
 	}
     }
     if (active->bf_len > 0) {
 	if (ioctl(ppp_fd, PPPIOCSACTIVE, active) < 0) {
-	    syslog(LOG_ERR, "Couldn't set active-filter in kernel: %m");
+	    error("Couldn't set active-filter in kernel: %m");
 	    ret = 0;
 	}
     }
@@ -847,17 +882,17 @@ sifvjcomp(u, vjcomp, cidcomp, maxcid)
     u_int x;
 
     if (ioctl(ppp_fd, PPPIOCGFLAGS, (caddr_t) &x) < 0) {
-	syslog(LOG_ERR, "ioctl (PPPIOCGFLAGS): %m");
+	error("ioctl (PPPIOCGFLAGS): %m");
 	return 0;
     }
     x = vjcomp ? x | SC_COMP_TCP: x &~ SC_COMP_TCP;
     x = cidcomp? x & ~SC_NO_TCP_CCID: x | SC_NO_TCP_CCID;
     if (ioctl(ppp_fd, PPPIOCSFLAGS, (caddr_t) &x) < 0) {
-	syslog(LOG_ERR, "ioctl(PPPIOCSFLAGS): %m");
+	error("ioctl(PPPIOCSFLAGS): %m");
 	return 0;
     }
     if (vjcomp && ioctl(ppp_fd, PPPIOCSMAXCID, (caddr_t) &maxcid) < 0) {
-	syslog(LOG_ERR, "ioctl(PPPIOCSFLAGS): %m");
+	error("ioctl(PPPIOCSFLAGS): %m");
 	return 0;
     }
     return 1;
@@ -872,14 +907,14 @@ sifup(u)
 {
     struct ifreq ifr;
 
-    strncpy(ifr.ifr_name, ifname, sizeof (ifr.ifr_name));
+    strlcpy(ifr.ifr_name, ifname, sizeof (ifr.ifr_name));
     if (ioctl(sockfd, SIOCGIFFLAGS, (caddr_t) &ifr) < 0) {
-	syslog(LOG_ERR, "ioctl (SIOCGIFFLAGS): %m");
+	error("ioctl (SIOCGIFFLAGS): %m");
 	return 0;
     }
     ifr.ifr_flags |= IFF_UP;
     if (ioctl(sockfd, SIOCSIFFLAGS, (caddr_t) &ifr) < 0) {
-	syslog(LOG_ERR, "ioctl(SIOCSIFFLAGS): %m");
+	error("ioctl(SIOCSIFFLAGS): %m");
 	return 0;
     }
     if_is_up = 1;
@@ -900,7 +935,7 @@ sifnpmode(u, proto, mode)
     npi.protocol = proto;
     npi.mode = mode;
     if (ioctl(ppp_fd, PPPIOCSNPMODE, &npi) < 0) {
-	syslog(LOG_ERR, "ioctl(set NP %d mode to %d): %m", proto, mode);
+	error("ioctl(set NP %d mode to %d): %m", proto, mode);
 	return 0;
     }
     return 1;
@@ -923,14 +958,14 @@ sifdown(u)
     ioctl(ppp_fd, PPPIOCSNPMODE, (caddr_t) &npi);
     /* ignore errors, because ppp_fd might have been closed by now. */
 
-    strncpy(ifr.ifr_name, ifname, sizeof (ifr.ifr_name));
+    strlcpy(ifr.ifr_name, ifname, sizeof (ifr.ifr_name));
     if (ioctl(sockfd, SIOCGIFFLAGS, (caddr_t) &ifr) < 0) {
-	syslog(LOG_ERR, "ioctl (SIOCGIFFLAGS): %m");
+	error("ioctl (SIOCGIFFLAGS): %m");
 	rv = 0;
     } else {
 	ifr.ifr_flags &= ~IFF_UP;
 	if (ioctl(sockfd, SIOCSIFFLAGS, (caddr_t) &ifr) < 0) {
-	    syslog(LOG_ERR, "ioctl(SIOCSIFFLAGS): %m");
+	    error("ioctl(SIOCSIFFLAGS): %m");
 	    rv = 0;
 	} else
 	    if_is_up = 0;
@@ -958,7 +993,7 @@ sifaddr(u, o, h, m)
     struct ifaliasreq ifra;
     struct ifreq ifr;
 
-    strncpy(ifra.ifra_name, ifname, sizeof(ifra.ifra_name));
+    strlcpy(ifra.ifra_name, ifname, sizeof(ifra.ifra_name));
     SET_SA_FAMILY(ifra.ifra_addr, AF_INET);
     ((struct sockaddr_in *) &ifra.ifra_addr)->sin_addr.s_addr = o;
     SET_SA_FAMILY(ifra.ifra_broadaddr, AF_INET);
@@ -969,19 +1004,17 @@ sifaddr(u, o, h, m)
     } else
 	BZERO(&ifra.ifra_mask, sizeof(ifra.ifra_mask));
     BZERO(&ifr, sizeof(ifr));
-    strncpy(ifr.ifr_name, ifname, sizeof(ifr.ifr_name));
+    strlcpy(ifr.ifr_name, ifname, sizeof(ifr.ifr_name));
     if (ioctl(sockfd, SIOCDIFADDR, (caddr_t) &ifr) < 0) {
 	if (errno != EADDRNOTAVAIL)
-	    syslog(LOG_WARNING, "Couldn't remove interface address: %m");
+	    warn("Couldn't remove interface address: %m");
     }
     if (ioctl(sockfd, SIOCAIFADDR, (caddr_t) &ifra) < 0) {
 	if (errno != EEXIST) {
-	    syslog(LOG_ERR, "Couldn't set interface address: %m");
+	    error("Couldn't set interface address: %m");
 	    return 0;
 	}
-	syslog(LOG_WARNING,
-	       "Couldn't set interface address: Address %s already exists",
-		ip_ntoa(o));
+	warn("Couldn't set interface address: Address %I already exists", o);
     }
     ifaddrs[0] = o;
     ifaddrs[1] = h;
@@ -1000,7 +1033,7 @@ cifaddr(u, o, h)
     struct ifaliasreq ifra;
 
     ifaddrs[0] = 0;
-    strncpy(ifra.ifra_name, ifname, sizeof(ifra.ifra_name));
+    strlcpy(ifra.ifra_name, ifname, sizeof(ifra.ifra_name));
     SET_SA_FAMILY(ifra.ifra_addr, AF_INET);
     ((struct sockaddr_in *) &ifra.ifra_addr)->sin_addr.s_addr = o;
     SET_SA_FAMILY(ifra.ifra_broadaddr, AF_INET);
@@ -1008,7 +1041,7 @@ cifaddr(u, o, h)
     BZERO(&ifra.ifra_mask, sizeof(ifra.ifra_mask));
     if (ioctl(sockfd, SIOCDIFADDR, (caddr_t) &ifra) < 0) {
 	if (errno != EADDRNOTAVAIL)
-	    syslog(LOG_WARNING, "Couldn't delete interface address: %m");
+	    warn("Couldn't delete interface address: %m");
 	return 0;
     }
     return 1;
@@ -1053,14 +1086,14 @@ dodefaultroute(g, cmd)
     } rtmsg;
 
     if ((routes = socket(PF_ROUTE, SOCK_RAW, AF_INET)) < 0) {
-	syslog(LOG_ERR, "Couldn't %s default route: socket: %m",
+	error("Couldn't %s default route: socket: %m",
 	       cmd=='s'? "add": "delete");
 	return 0;
     }
 
     memset(&rtmsg, 0, sizeof(rtmsg));
     rtmsg.hdr.rtm_type = cmd == 's'? RTM_ADD: RTM_DELETE;
-    rtmsg.hdr.rtm_flags = RTF_UP | RTF_GATEWAY;
+    rtmsg.hdr.rtm_flags = RTF_UP | RTF_GATEWAY | RTF_STATIC;
     rtmsg.hdr.rtm_version = RTM_VERSION;
     rtmsg.hdr.rtm_seq = ++rtm_seq;
     rtmsg.hdr.rtm_addrs = RTA_DST | RTA_GATEWAY | RTA_NETMASK;
@@ -1074,7 +1107,7 @@ dodefaultroute(g, cmd)
 
     rtmsg.hdr.rtm_msglen = sizeof(rtmsg);
     if (write(routes, &rtmsg, sizeof(rtmsg)) < 0) {
-	syslog(LOG_ERR, "Couldn't %s default route: %m",
+	error("Couldn't %s default route: %m",
 	       cmd=='s'? "add": "delete");
 	close(routes);
 	return 0;
@@ -1112,12 +1145,12 @@ sifproxyarp(unit, hisaddr)
      */
     memset(&arpmsg, 0, sizeof(arpmsg));
     if (!get_ether_addr(hisaddr, &arpmsg.hwa)) {
-	syslog(LOG_ERR, "Cannot determine ethernet address for proxy ARP");
+	error("Cannot determine ethernet address for proxy ARP");
 	return 0;
     }
 
     if ((routes = socket(PF_ROUTE, SOCK_RAW, AF_INET)) < 0) {
-	syslog(LOG_ERR, "Couldn't add proxy arp entry: socket: %m");
+	error("Couldn't add proxy arp entry: socket: %m");
 	return 0;
     }
 
@@ -1135,7 +1168,7 @@ sifproxyarp(unit, hisaddr)
     arpmsg.hdr.rtm_msglen = (char *) &arpmsg.hwa - (char *) &arpmsg
 	+ arpmsg.hwa.sdl_len;
     if (write(routes, &arpmsg, arpmsg.hdr.rtm_msglen) < 0) {
-	syslog(LOG_ERR, "Couldn't add proxy arp entry: %m");
+	error("Couldn't add proxy arp entry: %m");
 	close(routes);
 	return 0;
     }
@@ -1164,12 +1197,12 @@ cifproxyarp(unit, hisaddr)
     arpmsg.hdr.rtm_seq = ++rtm_seq;
 
     if ((routes = socket(PF_ROUTE, SOCK_RAW, AF_INET)) < 0) {
-	syslog(LOG_ERR, "Couldn't delete proxy arp entry: socket: %m");
+	error("Couldn't delete proxy arp entry: socket: %m");
 	return 0;
     }
 
     if (write(routes, &arpmsg, arpmsg.hdr.rtm_msglen) < 0) {
-	syslog(LOG_ERR, "Couldn't delete proxy arp entry: %m");
+	error("Couldn't delete proxy arp entry: %m");
 	close(routes);
 	return 0;
     }
@@ -1202,7 +1235,7 @@ sifproxyarp(unit, hisaddr)
      * as our local address.
      */
     if (!get_ether_addr(hisaddr, &dls.sdl)) {
-	syslog(LOG_ERR, "Cannot determine ethernet address for proxy ARP");
+	error("Cannot determine ethernet address for proxy ARP");
 	return 0;
     }
 
@@ -1213,7 +1246,7 @@ sifproxyarp(unit, hisaddr)
     ((struct sockaddr_in *) &arpreq.arp_pa)->sin_addr.s_addr = hisaddr;
     arpreq.arp_flags = ATF_PERM | ATF_PUBL;
     if (ioctl(sockfd, SIOCSARP, (caddr_t)&arpreq) < 0) {
-	syslog(LOG_ERR, "Couldn't add proxy arp entry: %m");
+	error("Couldn't add proxy arp entry: %m");
 	return 0;
     }
 
@@ -1235,7 +1268,7 @@ cifproxyarp(unit, hisaddr)
     SET_SA_FAMILY(arpreq.arp_pa, AF_INET);
     ((struct sockaddr_in *) &arpreq.arp_pa)->sin_addr.s_addr = hisaddr;
     if (ioctl(sockfd, SIOCDARP, (caddr_t)&arpreq) < 0) {
-	syslog(LOG_WARNING, "Couldn't delete proxy arp entry: %m");
+	warn("Couldn't delete proxy arp entry: %m");
 	return 0;
     }
     proxy_arp_addr = 0;
@@ -1265,7 +1298,7 @@ get_ether_addr(ipaddr, hwaddr)
     ifc.ifc_len = sizeof(ifs);
     ifc.ifc_req = ifs;
     if (ioctl(sockfd, SIOCGIFCONF, &ifc) < 0) {
-	syslog(LOG_ERR, "ioctl(SIOCGIFCONF): %m");
+	error("ioctl(SIOCGIFCONF): %m");
 	return 0;
     }
 
@@ -1278,7 +1311,7 @@ get_ether_addr(ipaddr, hwaddr)
 	 	((char *)&ifr->ifr_addr + ifr->ifr_addr.sa_len)) {
 	if (ifr->ifr_addr.sa_family == AF_INET) {
 	    ina = ((struct sockaddr_in *) &ifr->ifr_addr)->sin_addr.s_addr;
-	    strncpy(ifreq.ifr_name, ifr->ifr_name, sizeof(ifreq.ifr_name));
+	    strlcpy(ifreq.ifr_name, ifr->ifr_name, sizeof(ifreq.ifr_name));
 	    /*
 	     * Check that the interface is up, and not point-to-point
 	     * or loopback.
@@ -1304,7 +1337,7 @@ get_ether_addr(ipaddr, hwaddr)
 
     if (ifr >= ifend)
 	return 0;
-    syslog(LOG_INFO, "found interface %s for proxy arp", ifr->ifr_name);
+    info("found interface %s for proxy arp", ifr->ifr_name);
 
     /*
      * Now scan through again looking for a link-level address
@@ -1360,7 +1393,7 @@ GetMask(addr)
     ifc.ifc_len = sizeof(ifs);
     ifc.ifc_req = ifs;
     if (ioctl(sockfd, SIOCGIFCONF, &ifc) < 0) {
-	syslog(LOG_WARNING, "ioctl(SIOCGIFCONF): %m");
+	warn("ioctl(SIOCGIFCONF): %m");
 	return mask;
     }
     ifend = (struct ifreq *) (ifc.ifc_buf + ifc.ifc_len);
@@ -1377,7 +1410,7 @@ GetMask(addr)
 	/*
 	 * Check that the interface is up, and not point-to-point or loopback.
 	 */
-	strncpy(ifreq.ifr_name, ifr->ifr_name, sizeof(ifreq.ifr_name));
+	strlcpy(ifreq.ifr_name, ifr->ifr_name, sizeof(ifreq.ifr_name));
 	if (ioctl(sockfd, SIOCGIFFLAGS, &ifreq) < 0)
 	    continue;
 	if ((ifreq.ifr_flags & (IFF_UP|IFF_POINTOPOINT|IFF_LOOPBACK))
@@ -1394,6 +1427,17 @@ GetMask(addr)
     return mask;
 }
 
+/*
+ * have_route_to - determine if the system has any route to
+ * a given IP address.
+ * For demand mode to work properly, we have to ignore routes
+ * through our own interface.
+ */
+int have_route_to(u_int32_t addr)
+{
+    return -1;
+}
+
 /*
  * Use the hostid as part of the random number seed.
  */
@@ -1415,13 +1459,15 @@ lock(dev)
     char hdb_lock_buffer[12];
     int fd, pid, n;
     char *p;
+    size_t l;
 
     if ((p = strrchr(dev, '/')) != NULL)
 	dev = p + 1;
-    lock_file = malloc(strlen(LOCK_PREFIX) + strlen(dev) + 1);
+    l = strlen(LOCK_PREFIX) + strlen(dev) + 1;
+    lock_file = malloc(l);
     if (lock_file == NULL)
 	novm("lock file name");
-    strcat(strcpy(lock_file, LOCK_PREFIX), dev);
+    slprintf(lock_file, l, "%s%s", LOCK_PREFIX, dev);
 
     while ((fd = open(lock_file, O_EXCL | O_CREAT | O_RDWR, 0644)) < 0) {
 	if (errno == EEXIST
@@ -1429,7 +1475,7 @@ lock(dev)
 	    /* Read the lock file to find out who has the device locked */
 	    n = read(fd, hdb_lock_buffer, 11);
 	    if (n <= 0) {
-		syslog(LOG_ERR, "Can't read pid from lock file %s", lock_file);
+		error("Can't read pid from lock file %s", lock_file);
 		close(fd);
 	    } else {
 		hdb_lock_buffer[n] = 0;
@@ -1438,25 +1484,25 @@ lock(dev)
 		    /* pid no longer exists - remove the lock file */
 		    if (unlink(lock_file) == 0) {
 			close(fd);
-			syslog(LOG_NOTICE, "Removed stale lock on %s (pid %d)",
+			notice("Removed stale lock on %s (pid %d)",
 			       dev, pid);
 			continue;
 		    } else
-			syslog(LOG_WARNING, "Couldn't remove stale lock on %s",
+			warn("Couldn't remove stale lock on %s",
 			       dev);
 		} else
-		    syslog(LOG_NOTICE, "Device %s is locked by pid %d",
+		    notice("Device %s is locked by pid %d",
 			   dev, pid);
 	    }
 	    close(fd);
 	} else
-	    syslog(LOG_ERR, "Can't create lock file %s: %m", lock_file);
+	    error("Can't create lock file %s: %m", lock_file);
 	free(lock_file);
 	lock_file = NULL;
 	return -1;
     }
 
-    sprintf(hdb_lock_buffer, "%10d\n", getpid());
+    slprintf(hdb_lock_buffer, sizeof(hdb_lock_buffer), "%10d\n", getpid());
     write(fd, hdb_lock_buffer, 11);
 
     close(fd);
