@@ -26,7 +26,7 @@
  */
 
 #ifndef lint
-static char rcsid[] = "$Id: sys-svr4.c,v 1.17 1998/03/25 02:19:31 paulus Exp $";
+static char rcsid[] = "$Id: sys-svr4.c,v 1.29 1999/03/22 05:55:39 paulus Exp $";
 #endif
 
 #include <limits.h>
@@ -49,7 +49,6 @@ static char rcsid[] = "$Id: sys-svr4.c,v 1.17 1998/03/25 02:19:31 paulus Exp $";
 #include <sys/stropts.h>
 #include <sys/socket.h>
 #include <sys/sockio.h>
-#include <sys/syslog.h>
 #include <sys/sysmacros.h>
 #include <sys/systeminfo.h>
 #include <sys/dlpi.h>
@@ -61,6 +60,10 @@ static char rcsid[] = "$Id: sys-svr4.c,v 1.17 1998/03/25 02:19:31 paulus Exp $";
 #include <net/ppp_defs.h>
 #include <net/pppio.h>
 #include <netinet/in.h>
+#include <sys/tihdr.h>
+#include <sys/tiuser.h>
+#include <inet/common.h>
+#include <inet/mib2.h>
 
 #include "pppd.h"
 
@@ -80,13 +83,19 @@ static pid_t	tty_sid;	/* original session ID for terminal */
 
 extern u_char	inpacket_buf[];	/* borrowed from main.c */
 
+#define MAX_POLLFDS	32
+static struct pollfd pollfds[MAX_POLLFDS];
+static int n_pollfds;
+
 static int	link_mtu, link_mru;
 
 #define NMODULES	32
 static int	tty_nmodules;
 static char	tty_modules[NMODULES][FMNAMESZ+1];
+static int	tty_npushed;
 
 static int	if_is_up;	/* Interface has been marked up */
+static u_int32_t remote_addr;		/* IP address of peer */
 static u_int32_t default_route_gateway;	/* Gateway for default route added */
 static u_int32_t proxy_arp_addr;	/* Addr for proxy arp entry added */
 
@@ -94,7 +103,7 @@ static u_int32_t proxy_arp_addr;	/* Addr for proxy arp entry added */
 static int translate_speed __P((int));
 static int baud_rate_of __P((int));
 static int get_ether_addr __P((u_int32_t, struct sockaddr *));
-static int get_hw_addr __P((char *, struct sockaddr *));
+static int get_hw_addr __P((char *, u_int32_t, struct sockaddr *));
 static int dlpi_attach __P((int, int));
 static int dlpi_info_req __P((int));
 static int dlpi_get_reply __P((int, union DL_primitives *, int, int));
@@ -117,29 +126,23 @@ sys_init()
 #endif
 
     ipfd = open("/dev/ip", O_RDWR, 0);
-    if (ipfd < 0) {
-	syslog(LOG_ERR, "Couldn't open IP device: %m");
-	die(1);
-    }
+    if (ipfd < 0)
+	fatal("Couldn't open IP device: %m");
 
-    if (default_device)
+    if (default_device && !notty)
 	tty_sid = getsid((pid_t)0);
 
     pppfd = open("/dev/ppp", O_RDWR | O_NONBLOCK, 0);
-    if (pppfd < 0) {
-	syslog(LOG_ERR, "Can't open /dev/ppp: %m");
-	die(1);
-    }
+    if (pppfd < 0)
+	fatal("Can't open /dev/ppp: %m");
     if (kdebugflag & 1) {
 	x = PPPDBG_LOG + PPPDBG_DRIVER;
 	strioctl(pppfd, PPPIO_DEBUG, &x, sizeof(int), 0);
     }
 
     /* Assign a new PPA and get its unit number. */
-    if (strioctl(pppfd, PPPIO_NEWPPA, &ifunit, 0, sizeof(int)) < 0) {
-	syslog(LOG_ERR, "Can't create new PPP interface: %m");
-	die(1);
-    }
+    if (strioctl(pppfd, PPPIO_NEWPPA, &ifunit, 0, sizeof(int)) < 0)
+	fatal("Can't create new PPP interface: %m");
 
     /*
      * Open the ppp device again and link it under the ip multiplexor.
@@ -147,44 +150,38 @@ sys_init()
      * I don't know any way to be certain they will be the same. :-(
      */
     ifd = open("/dev/ppp", O_RDWR, 0);
-    if (ifd < 0) {
-	syslog(LOG_ERR, "Can't open /dev/ppp (2): %m");
-	die(1);
-    }
+    if (ifd < 0)
+	fatal("Can't open /dev/ppp (2): %m");
     if (kdebugflag & 1) {
 	x = PPPDBG_LOG + PPPDBG_DRIVER;
 	strioctl(ifd, PPPIO_DEBUG, &x, sizeof(int), 0);
     }
 #ifdef sun
     if (ioctl(ifd, I_PUSH, "ip") < 0) {
-	syslog(LOG_ERR, "Can't push IP module: %m");
 	close(ifd);
-	die(1);
+	fatal("Can't push IP module: %m");
     }
 #else
     if (dlpi_attach(ifd, ifunit) < 0 ||
 	dlpi_get_reply(ifd, &reply.prim, DL_OK_ACK, sizeof(reply)) < 0) {
-	syslog(LOG_ERR, "Can't attach to ppp%d: %m", ifunit);
 	close(ifd);
-	die(1);
+	fatal("Can't attach to ppp%d: %m", ifunit);
     }
 #endif
     ipmuxid = ioctl(ipfd, I_LINK, ifd);
     close(ifd);
-    if (ipmuxid < 0) {
-	syslog(LOG_ERR, "Can't link PPP device to IP: %m");
-	die(1);
-    }
+    if (ipmuxid < 0)
+	fatal("Can't link PPP device to IP: %m");
 
 #ifndef sun
     /* Set the interface name for the link. */
-    (void) sprintf (ifr.ifr_name, "ppp%d", ifunit);
+    slprintf(ifr.ifr_name, sizeof(ifr.ifr_name), "ppp%d", ifunit);
     ifr.ifr_metric = ipmuxid;
-    if (strioctl(ipfd, SIOCSIFNAME, (char *)&ifr, sizeof ifr, 0) < 0) {
-	syslog(LOG_ERR, "Can't set interface name %s: %m", ifr.ifr_name);
-	die(1);
-    }
+    if (strioctl(ipfd, SIOCSIFNAME, (char *)&ifr, sizeof ifr, 0) < 0)
+	fatal("Can't set interface name %s: %m", ifr.ifr_name);
 #endif
+
+    n_pollfds = 0;
 }
 
 /*
@@ -219,9 +216,10 @@ sys_close()
 /*
  * sys_check_options - check the options that the user specified
  */
-void
+int
 sys_check_options()
 {
+    return 1;
 }
 
 
@@ -263,7 +261,7 @@ ppp_available()
 /*
  * establish_ppp - Turn the serial port into a ppp interface.
  */
-void
+int
 establish_ppp(fd)
     int fd;
 {
@@ -272,23 +270,26 @@ establish_ppp(fd)
     /* Pop any existing modules off the tty stream. */
     for (i = 0;; ++i)
 	if (ioctl(fd, I_LOOK, tty_modules[i]) < 0
+	    || strcmp(tty_modules[i], "ptem") == 0
 	    || ioctl(fd, I_POP, 0) < 0)
 	    break;
     tty_nmodules = i;
 
     /* Push the async hdlc module and the compressor module. */
+    tty_npushed = 0;
     if (ioctl(fd, I_PUSH, "ppp_ahdl") < 0) {
-	syslog(LOG_ERR, "Couldn't push PPP Async HDLC module: %m");
-	die(1);
+	error("Couldn't push PPP Async HDLC module: %m");
+	return -1;
     }
+    ++tty_npushed;
     if (kdebugflag & 4) {
 	i = PPPDBG_LOG + PPPDBG_AHDLC;
 	strioctl(pppfd, PPPIO_DEBUG, &i, sizeof(int), 0);
     }
-    if (ioctl(fd, I_PUSH, "ppp_comp") < 0) {
-	syslog(LOG_ERR, "Couldn't push PPP compression module: %m");
-/*	die(1); */
-    }
+    if (ioctl(fd, I_PUSH, "ppp_comp") < 0)
+	error("Couldn't push PPP compression module: %m");
+    else
+	++tty_npushed;
     if (kdebugflag & 2) {
 	i = PPPDBG_LOG + PPPDBG_COMP;
 	strioctl(pppfd, PPPIO_DEBUG, &i, sizeof(int), 0);
@@ -296,9 +297,11 @@ establish_ppp(fd)
 
     /* Link the serial port under the PPP multiplexor. */
     if ((fdmuxid = ioctl(pppfd, I_LINK, fd)) < 0) {
-	syslog(LOG_ERR, "Can't link tty to PPP mux: %m");
-	die(1);
+	error("Can't link tty to PPP mux: %m");
+	return -1;
     }
+
+    return pppfd;
 }
 
 /*
@@ -324,16 +327,16 @@ disestablish_ppp(fd)
     if (fdmuxid >= 0) {
 	if (ioctl(pppfd, I_UNLINK, fdmuxid) < 0) {
 	    if (!hungup)
-		syslog(LOG_ERR, "Can't unlink tty from PPP mux: %m");
+		error("Can't unlink tty from PPP mux: %m");
 	}
 	fdmuxid = -1;
 
 	if (!hungup) {
-	    while (ioctl(fd, I_POP, 0) >= 0)
-		;
+	    while (tty_npushed > 0 && ioctl(fd, I_POP, 0) >= 0)
+		--tty_npushed;
 	    for (i = tty_nmodules - 1; i >= 0; --i)
 		if (ioctl(fd, I_PUSH, tty_modules[i]) < 0)
-		    syslog(LOG_ERR, "Couldn't restore tty module %s: %m",
+		    error("Couldn't restore tty module %s: %m",
 			   tty_modules[i]);
 	}
 	if (hungup && default_device && tty_sid > 0) {
@@ -376,8 +379,8 @@ clean_check()
 	break;
     }
     if (s != NULL) {
-	syslog(LOG_WARNING, "Serial link is not 8-bit clean:");
-	syslog(LOG_WARNING, "All received characters had %s", s);
+	warn("Serial link is not 8-bit clean:");
+	warn("All received characters had %s", s);
     }
 }
 
@@ -470,7 +473,7 @@ translate_speed(bps)
     for (speedp = speeds; speedp->speed_int; speedp++)
 	if (bps == speedp->speed_int)
 	    return speedp->speed_val;
-    syslog(LOG_WARNING, "speed %d not supported", bps);
+    warn("speed %d not supported", bps);
     return 0;
 }
 
@@ -506,17 +509,15 @@ set_up_tty(fd, local)
     struct termiox tiox;
 #endif
 
-    if (tcgetattr(fd, &tios) < 0) {
-	syslog(LOG_ERR, "tcgetattr: %m");
-	die(1);
-    }
+    if (tcgetattr(fd, &tios) < 0)
+	fatal("tcgetattr: %m");
 
 #ifndef CRTSCTS
     termiox_ok = 1;
     if (ioctl (fd, TCGETX, &tiox) < 0) {
 	termiox_ok = 0;
 	if (errno != ENOTTY)
-	    syslog (LOG_ERR, "TCGETX: %m");
+	    error("TCGETX: %m");
     }
 #endif
 
@@ -536,7 +537,7 @@ set_up_tty(fd, local)
 	tios.c_cflag &= ~CRTSCTS;
 #else
     if (crtscts != 0 && !termiox_ok) {
-	syslog(LOG_ERR, "Can't set RTS/CTS flow control");
+	error("Can't set RTS/CTS flow control");
     } else if (crtscts > 0) {
 	tiox.x_hflag |= RTSXOFF|CTSXON;
     } else if (crtscts < 0) {
@@ -569,21 +570,16 @@ set_up_tty(fd, local)
 	 * We can't proceed if the serial port speed is 0,
 	 * since that implies that the serial port is disabled.
 	 */
-	if (speed == B0) {
-	    syslog(LOG_ERR, "Baud rate for %s is 0; need explicit baud rate",
-		   devnam);
-	    die(1);
-	}
+	if (speed == B0)
+	    fatal("Baud rate for %s is 0; need explicit baud rate", devnam);
     }
 
-    if (tcsetattr(fd, TCSAFLUSH, &tios) < 0) {
-	syslog(LOG_ERR, "tcsetattr: %m");
-	die(1);
-    }
+    if (tcsetattr(fd, TCSAFLUSH, &tios) < 0)
+	fatal("tcsetattr: %m");
 
 #ifndef CRTSCTS
     if (termiox_ok && ioctl (fd, TCSETXF, &tiox) < 0){
-	syslog (LOG_ERR, "TCSETXF: %m");
+	error("TCSETXF: %m");
     }
 #endif
 
@@ -610,11 +606,11 @@ restore_tty(fd)
 	}
 	if (tcsetattr(fd, TCSAFLUSH, &inittermios) < 0)
 	    if (!hungup && errno != ENXIO)
-		syslog(LOG_WARNING, "tcsetattr: %m");
+		warn("tcsetattr: %m");
 #ifndef CRTSCTS
 	if (ioctl (fd, TCSETXF, &inittermiox) < 0){
 	    if (!hungup && errno != ENXIO)
-		syslog (LOG_ERR, "TCSETXF: %m");
+		error("TCSETXF: %m");
 	}
 #endif
 	ioctl(fd, TIOCSWINSZ, &wsinfo);
@@ -640,9 +636,10 @@ int fd, on;
  * in demand mode.  Under Solaris 2, we use our existing fd
  * to the ppp driver.
  */
-void
+int
 open_ppp_loopback()
 {
+    return pppfd;
 }
 
 /*
@@ -659,7 +656,7 @@ output(unit, p, len)
     struct pollfd pfd;
 
     if (debug)
-	log_packet(p, len, "sent ", LOG_DEBUG);
+	dbglog("sent %P", p, len);
 
     data.len = len;
     data.buf = (caddr_t) p;
@@ -667,7 +664,7 @@ output(unit, p, len)
     while (putmsg(pppfd, NULL, &data, 0) < 0) {
 	if (--retries < 0 || (errno != EWOULDBLOCK && errno != EAGAIN)) {
 	    if (errno != ENXIO)
-		syslog(LOG_ERR, "Couldn't send packet: %m");
+		error("Couldn't send packet: %m");
 	    break;
 	}
 	pfd.fd = pppfd;
@@ -678,7 +675,7 @@ output(unit, p, len)
 
 
 /*
- * wait_input - wait until there is data available on fd,
+ * wait_input - wait until there is data available,
  * for the length of time specified by *timo (indefinite
  * if timo is NULL).
  */
@@ -687,17 +684,50 @@ wait_input(timo)
     struct timeval *timo;
 {
     int t;
-    struct pollfd pfd;
 
     t = timo == NULL? -1: timo->tv_sec * 1000 + timo->tv_usec / 1000;
-    pfd.fd = pppfd;
-    pfd.events = POLLIN | POLLPRI | POLLHUP;
-    if (poll(&pfd, 1, t) < 0 && errno != EINTR) {
-	syslog(LOG_ERR, "poll: %m");
-	die(1);
+    if (poll(pollfds, n_pollfds, t) < 0 && errno != EINTR)
+	fatal("poll: %m");
+}
+
+/*
+ * add_fd - add an fd to the set that wait_input waits for.
+ */
+void add_fd(fd)
+    int fd;
+{
+    int n;
+
+    for (n = 0; n < n_pollfds; ++n)
+	if (pollfds[n].fd == fd)
+	    return;
+    if (n_pollfds < MAX_POLLFDS) {
+	pollfds[n_pollfds].fd = fd;
+	pollfds[n_pollfds].events = POLLIN | POLLPRI | POLLHUP;
+	++n_pollfds;
+    } else
+	error("Too many inputs!");
+}
+
+/*
+ * remove_fd - remove an fd from the set that wait_input waits for.
+ */
+void remove_fd(fd)
+    int fd;
+{
+    int n;
+
+    for (n = 0; n < n_pollfds; ++n) {
+	if (pollfds[n].fd == fd) {
+	    while (++n < n_pollfds)
+		pollfds[n-1] = pollfds[n];
+	    --n_pollfds;
+	    break;
+	}
     }
 }
 
+#if 0
 /*
  * wait_loop_output - wait until there is data available on the
  * loopback, for the length of time specified by *timo (indefinite
@@ -721,11 +751,10 @@ wait_time(timo)
     int n;
 
     n = select(0, NULL, NULL, NULL, timo);
-    if (n < 0 && errno != EINTR) {
-	syslog(LOG_ERR, "select: %m");
-	die(1);
-    }
+    if (n < 0 && errno != EINTR)
+	fatal("select: %m");
 }
+#endif
 
 
 /*
@@ -747,10 +776,9 @@ read_packet(buf)
 	flags = 0;
 	len = getmsg(pppfd, &ctrl, &data, &flags);
 	if (len < 0) {
-	    if (errno = EAGAIN || errno == EINTR)
+	    if (errno == EAGAIN || errno == EINTR)
 		return -1;
-	    syslog(LOG_ERR, "Error reading packet: %m");
-	    die(1);
+	    fatal("Error reading packet: %m");
 	}
 
 	if (ctrl.len <= 0)
@@ -761,7 +789,7 @@ read_packet(buf)
 	 * as a DLPI primitive??
 	 */
 	if (debug)
-	    syslog(LOG_DEBUG, "got dlpi prim 0x%x, len=%d",
+	    dbglog("got dlpi prim 0x%x, len=%d",
 		   ((union DL_primitives *)ctrlbuf)->dl_primitive, ctrl.len);
 
     }
@@ -802,26 +830,26 @@ ppp_send_config(unit, mtu, asyncmap, pcomp, accomp)
     if (strioctl(pppfd, PPPIO_MTU, &mtu, sizeof(mtu), 0) < 0) {
 	if (hungup && errno == ENXIO)
 	    return;
-	syslog(LOG_ERR, "Couldn't set MTU: %m");
+	error("Couldn't set MTU: %m");
     }
     if (fdmuxid >= 0) {
 	/* can't set these if we don't have a stream attached below /dev/ppp */
 	if (strioctl(pppfd, PPPIO_XACCM, &asyncmap, sizeof(asyncmap), 0) < 0) {
-	    syslog(LOG_ERR, "Couldn't set transmit ACCM: %m");
+	    error("Couldn't set transmit ACCM: %m");
 	}
 	cf[0] = (pcomp? COMP_PROT: 0) + (accomp? COMP_AC: 0);
 	cf[1] = COMP_PROT | COMP_AC;
 	if (strioctl(pppfd, PPPIO_CFLAGS, cf, sizeof(cf), sizeof(int)) < 0) {
-	    syslog(LOG_ERR, "Couldn't set prot/AC compression: %m");
+	    error("Couldn't set prot/AC compression: %m");
 	}
     }
 
     /* set the MTU for IP as well */
     memset(&ifr, 0, sizeof(ifr));
-    strncpy(ifr.ifr_name, ifname, sizeof(ifr.ifr_name));
+    strlcpy(ifr.ifr_name, ifname, sizeof(ifr.ifr_name));
     ifr.ifr_metric = link_mtu;
     if (ioctl(ipfd, SIOCSIFMTU, &ifr) < 0) {
-	syslog(LOG_ERR, "Couldn't set IP MTU: %m");
+	error("Couldn't set IP MTU: %m");
     }
 }
 
@@ -836,7 +864,7 @@ ppp_set_xaccm(unit, accm)
     if (fdmuxid >= 0
 	&& strioctl(pppfd, PPPIO_XACCM, accm, sizeof(ext_accm), 0) < 0) {
 	if (!hungup || errno != ENXIO)
-	    syslog(LOG_WARNING, "Couldn't set extended ACCM: %m");
+	    warn("Couldn't set extended ACCM: %m");
     }
 }
 
@@ -856,17 +884,17 @@ ppp_recv_config(unit, mru, asyncmap, pcomp, accomp)
     if (strioctl(pppfd, PPPIO_MRU, &mru, sizeof(mru), 0) < 0) {
 	if (hungup && errno == ENXIO)
 	    return;
-	syslog(LOG_ERR, "Couldn't set MRU: %m");
+	error("Couldn't set MRU: %m");
     }
     if (fdmuxid >= 0) {
 	/* can't set these if we don't have a stream attached below /dev/ppp */
 	if (strioctl(pppfd, PPPIO_RACCM, &asyncmap, sizeof(asyncmap), 0) < 0) {
-	    syslog(LOG_ERR, "Couldn't set receive ACCM: %m");
+	    error("Couldn't set receive ACCM: %m");
 	}
 	cf[0] = (pcomp? DECOMP_PROT: 0) + (accomp? DECOMP_AC: 0);
 	cf[1] = DECOMP_PROT | DECOMP_AC;
 	if (strioctl(pppfd, PPPIO_CFLAGS, cf, sizeof(cf), sizeof(int)) < 0) {
-	    syslog(LOG_ERR, "Couldn't set prot/AC decompression: %m");
+	    error("Couldn't set prot/AC decompression: %m");
 	}
     }
 }
@@ -899,7 +927,7 @@ ccp_flags_set(unit, isopen, isup)
     cf[1] = CCP_ISOPEN | CCP_ISUP | CCP_ERROR | CCP_FATALERROR;
     if (strioctl(pppfd, PPPIO_CFLAGS, cf, sizeof(cf), sizeof(int)) < 0) {
 	if (!hungup || errno != ENXIO)
-	    syslog(LOG_ERR, "Couldn't set kernel CCP state: %m");
+	    error("Couldn't set kernel CCP state: %m");
     }
 }
 
@@ -914,6 +942,25 @@ get_idle_time(u, ip)
     return strioctl(pppfd, PPPIO_GIDLE, ip, 0, sizeof(struct ppp_idle)) >= 0;
 }
 
+/*
+ * get_ppp_stats - return statistics for the link.
+ */
+int
+get_ppp_stats(u, stats)
+    int u;
+    struct pppd_stats *stats;
+{
+    struct ppp_stats s;
+
+    if (strioctl(pppfd, PPPIO_GETSTAT, &s, 0, sizeof(s)) < 0) {
+	error("Couldn't get link statistics: %m");
+	return 0;
+    }
+    stats->bytes_in = s.p.ppp_ibytes;
+    stats->bytes_out = s.p.ppp_obytes;
+    return 1;
+}
+
 #if 0
 /*
  * set_filters - transfer the pass and active filters to the kernel.
@@ -927,14 +974,14 @@ set_filters(pass, active)
     if (pass->bf_len > 0) {
 	if (strioctl(pppfd, PPPIO_PASSFILT, pass,
 		     sizeof(struct bpf_program), 0) < 0) {
-	    syslog(LOG_ERR, "Couldn't set pass-filter in kernel: %m");
+	    error("Couldn't set pass-filter in kernel: %m");
 	    ret = 0;
 	}
     }
     if (active->bf_len > 0) {
 	if (strioctl(pppfd, PPPIO_ACTIVEFILT, active,
 		     sizeof(struct bpf_program), 0) < 0) {
-	    syslog(LOG_ERR, "Couldn't set active-filter in kernel: %m");
+	    error("Couldn't set active-filter in kernel: %m");
 	    ret = 0;
 	}
     }
@@ -956,7 +1003,7 @@ ccp_fatal_error(unit)
     cf[0] = cf[1] = 0;
     if (strioctl(pppfd, PPPIO_CFLAGS, cf, sizeof(cf), sizeof(int)) < 0) {
 	if (errno != ENXIO && errno != EINVAL)
-	    syslog(LOG_ERR, "Couldn't get compression flags: %m");
+	    error("Couldn't get compression flags: %m");
 	return 0;
     }
     return cf[0] & CCP_FATALERROR;
@@ -976,7 +1023,7 @@ sifvjcomp(u, vjcomp, xcidcomp, xmaxcid)
 	maxcid[0] = xcidcomp;
 	maxcid[1] = 15;		/* XXX should be rmaxcid */
 	if (strioctl(pppfd, PPPIO_VJINIT, maxcid, sizeof(maxcid), 0) < 0) {
-	    syslog(LOG_ERR, "Couldn't initialize VJ compression: %m");
+	    error("Couldn't initialize VJ compression: %m");
 	}
     }
 
@@ -985,7 +1032,7 @@ sifvjcomp(u, vjcomp, xcidcomp, xmaxcid)
     cf[1] = COMP_VJC + DECOMP_VJC + COMP_VJCCID + DECOMP_VJCCID;
     if (strioctl(pppfd, PPPIO_CFLAGS, cf, sizeof(cf), sizeof(int)) < 0) {
 	if (vjcomp)
-	    syslog(LOG_ERR, "Couldn't enable VJ compression: %m");
+	    error("Couldn't enable VJ compression: %m");
     }
 
     return 1;
@@ -1000,14 +1047,14 @@ sifup(u)
 {
     struct ifreq ifr;
 
-    strncpy(ifr.ifr_name, ifname, sizeof(ifr.ifr_name));
+    strlcpy(ifr.ifr_name, ifname, sizeof(ifr.ifr_name));
     if (ioctl(ipfd, SIOCGIFFLAGS, &ifr) < 0) {
-	syslog(LOG_ERR, "Couldn't mark interface up (get): %m");
+	error("Couldn't mark interface up (get): %m");
 	return 0;
     }
     ifr.ifr_flags |= IFF_UP;
     if (ioctl(ipfd, SIOCSIFFLAGS, &ifr) < 0) {
-	syslog(LOG_ERR, "Couldn't mark interface up (set): %m");
+	error("Couldn't mark interface up (set): %m");
 	return 0;
     }
     if_is_up = 1;
@@ -1025,14 +1072,14 @@ sifdown(u)
 
     if (ipmuxid < 0)
 	return 1;
-    strncpy(ifr.ifr_name, ifname, sizeof(ifr.ifr_name));
+    strlcpy(ifr.ifr_name, ifname, sizeof(ifr.ifr_name));
     if (ioctl(ipfd, SIOCGIFFLAGS, &ifr) < 0) {
-	syslog(LOG_ERR, "Couldn't mark interface down (get): %m");
+	error("Couldn't mark interface down (get): %m");
 	return 0;
     }
     ifr.ifr_flags &= ~IFF_UP;
     if (ioctl(ipfd, SIOCSIFFLAGS, &ifr) < 0) {
-	syslog(LOG_ERR, "Couldn't mark interface down (set): %m");
+	error("Couldn't mark interface down (set): %m");
 	return 0;
     }
     if_is_up = 0;
@@ -1053,7 +1100,7 @@ sifnpmode(u, proto, mode)
     npi[0] = proto;
     npi[1] = (int) mode;
     if (strioctl(pppfd, PPPIO_NPMODE, &npi, 2 * sizeof(int), 0) < 0) {
-	syslog(LOG_ERR, "ioctl(set NP %d mode to %d): %m", proto, mode);
+	error("ioctl(set NP %d mode to %d): %m", proto, mode);
 	return 0;
     }
     return 1;
@@ -1073,17 +1120,17 @@ sifaddr(u, o, h, m)
     int ret = 1;
 
     memset(&ifr, 0, sizeof(ifr));
-    strncpy(ifr.ifr_name, ifname, sizeof(ifr.ifr_name));
+    strlcpy(ifr.ifr_name, ifname, sizeof(ifr.ifr_name));
     ifr.ifr_addr.sa_family = AF_INET;
     INET_ADDR(ifr.ifr_addr) = m;
     if (ioctl(ipfd, SIOCSIFNETMASK, &ifr) < 0) {
-	syslog(LOG_ERR, "Couldn't set IP netmask: %m");
+	error("Couldn't set IP netmask: %m");
 	ret = 0;
     }
     ifr.ifr_addr.sa_family = AF_INET;
     INET_ADDR(ifr.ifr_addr) = o;
     if (ioctl(ipfd, SIOCSIFADDR, &ifr) < 0) {
-	syslog(LOG_ERR, "Couldn't set local IP address: %m");
+	error("Couldn't set local IP address: %m");
 	ret = 0;
     }
 
@@ -1095,23 +1142,24 @@ sifaddr(u, o, h, m)
 	&& (ifr.ifr_flags & IFF_POINTOPOINT) == 0) {
 	ifr.ifr_flags |= IFF_POINTOPOINT;
 	if (ioctl(ipfd, SIOCSIFFLAGS, &ifr) < 0) {
-	    syslog(LOG_ERR, "Couldn't mark interface pt-to-pt: %m");
+	    error("Couldn't mark interface pt-to-pt: %m");
 	    ret = 0;
 	}
     }
     ifr.ifr_dstaddr.sa_family = AF_INET;
     INET_ADDR(ifr.ifr_dstaddr) = h;
     if (ioctl(ipfd, SIOCSIFDSTADDR, &ifr) < 0) {
-	syslog(LOG_ERR, "Couldn't set remote IP address: %m");
+	error("Couldn't set remote IP address: %m");
 	ret = 0;
     }
 #if 0	/* now done in ppp_send_config */
     ifr.ifr_metric = link_mtu;
     if (ioctl(ipfd, SIOCSIFMTU, &ifr) < 0) {
-	syslog(LOG_ERR, "Couldn't set IP MTU: %m");
+	error("Couldn't set IP MTU: %m");
     }
 #endif
 
+    remote_addr = h;
     return ret;
 }
 
@@ -1127,14 +1175,15 @@ cifaddr(u, o, h)
 #if defined(__USLC__)		/* was: #if 0 */
     cifroute(unit, ouraddr, hisaddr);
     if (ipmuxid >= 0) {
-	syslog(LOG_NOTICE, "Removing ppp interface unit");
+	notice("Removing ppp interface unit");
 	if (ioctl(ipfd, I_UNLINK, ipmuxid) < 0) {
-	    syslog(LOG_ERR, "Can't remove ppp interface unit: %m");
+	    error("Can't remove ppp interface unit: %m");
 	    return 0;
 	}
 	ipmuxid = -1;
     }
 #endif
+    remote_addr = 0;
     return 1;
 }
 
@@ -1159,7 +1208,7 @@ sifdefaultroute(u, l, g)
     rt.rt_flags = RTF_GATEWAY;
 
     if (ioctl(ipfd, SIOCADDRT, &rt) < 0) {
-	syslog(LOG_ERR, "Can't add default route: %m");
+	error("Can't add default route: %m");
 	return 0;
     }
 
@@ -1188,7 +1237,7 @@ cifdefaultroute(u, l, g)
     rt.rt_flags = RTF_GATEWAY;
 
     if (ioctl(ipfd, SIOCDELRT, &rt) < 0) {
-	syslog(LOG_ERR, "Can't delete default route: %m");
+	error("Can't delete default route: %m");
 	return 0;
     }
 
@@ -1214,7 +1263,7 @@ sifproxyarp(unit, hisaddr)
     INET_ADDR(arpreq.arp_pa) = hisaddr;
     arpreq.arp_flags = ATF_PERM | ATF_PUBL;
     if (ioctl(ipfd, SIOCSARP, (caddr_t) &arpreq) < 0) {
-	syslog(LOG_ERR, "Couldn't set proxy ARP entry: %m");
+	error("Couldn't set proxy ARP entry: %m");
 	return 0;
     }
 
@@ -1236,7 +1285,7 @@ cifproxyarp(unit, hisaddr)
     arpreq.arp_pa.sa_family = AF_INET;
     INET_ADDR(arpreq.arp_pa) = hisaddr;
     if (ioctl(ipfd, SIOCDARP, (caddr_t)&arpreq) < 0) {
-	syslog(LOG_ERR, "Couldn't delete proxy ARP entry: %m");
+	error("Couldn't delete proxy ARP entry: %m");
 	return 0;
     }
 
@@ -1272,7 +1321,7 @@ get_ether_addr(ipaddr, hwaddr)
     if (ifc.ifc_buf == 0)
 	return 0;
     if (ioctl(ipfd, SIOCGIFCONF, &ifc) < 0) {
-	syslog(LOG_WARNING, "Couldn't get system interface list: %m");
+	warn("Couldn't get system interface list: %m");
 	free(ifc.ifc_buf);
 	return 0;
     }
@@ -1283,7 +1332,7 @@ get_ether_addr(ipaddr, hwaddr)
 	/*
 	 * Check that the interface is up, and not point-to-point or loopback.
 	 */
-	strncpy(ifreq.ifr_name, ifr->ifr_name, sizeof(ifreq.ifr_name));
+	strlcpy(ifreq.ifr_name, ifr->ifr_name, sizeof(ifreq.ifr_name));
 	if (ioctl(ipfd, SIOCGIFFLAGS, &ifreq) < 0)
 	    continue;
 	if ((ifreq.ifr_flags &
@@ -1302,14 +1351,14 @@ get_ether_addr(ipaddr, hwaddr)
     }
 
     if (ifr >= ifend) {
-	syslog(LOG_WARNING, "No suitable interface found for proxy ARP");
+	warn("No suitable interface found for proxy ARP");
 	free(ifc.ifc_buf);
 	return 0;
     }
 
-    syslog(LOG_INFO, "found interface %s for proxy ARP", ifr->ifr_name);
-    if (!get_hw_addr(ifr->ifr_name, hwaddr)) {
-	syslog(LOG_ERR, "Couldn't get hardware address for %s", ifr->ifr_name);
+    info("found interface %s for proxy ARP", ifr->ifr_name);
+    if (!get_hw_addr(ifr->ifr_name, ina, hwaddr)) {
+	error("Couldn't get hardware address for %s", ifr->ifr_name);
 	free(ifc.ifc_buf);
 	return 0;
     }
@@ -1322,10 +1371,30 @@ get_ether_addr(ipaddr, hwaddr)
  * get_hw_addr - obtain the hardware address for a named interface.
  */
 static int
-get_hw_addr(name, hwaddr)
+get_hw_addr(name, ina, hwaddr)
     char *name;
+    u_int32_t ina;
     struct sockaddr *hwaddr;
 {
+#if 1
+    /* New way - get the address by doing an arp request. */
+    int s;
+    struct arpreq req;
+
+    s = socket(AF_INET, SOCK_DGRAM, 0);
+    if (s < 0)
+	return 0;
+    memset(&req, 0, sizeof(req));
+    req.arp_pa.sa_family = AF_INET;
+    INET_ADDR(req.arp_pa) = ina;
+    if (ioctl(s, SIOCGARP, &req) < 0) {
+	error("Couldn't get ARP entry for %s: %m", ip_ntoa(ina));
+	return 0;
+    }
+    *hwaddr = req.arp_ha;
+    hwaddr->sa_family = AF_UNSPEC;
+
+#else /* 0 */
     char *p, *q;
     int unit, iffd, adrlen;
     unsigned char *adrp;
@@ -1339,19 +1408,19 @@ get_hw_addr(name, hwaddr)
      * We have to open the device and ask it for its hardware address.
      * First split apart the device name and unit.
      */
-    strcpy(ifdev, "/dev/");
-    q = ifdev + 5;		/* strlen("/dev/") */
-    while (*name != 0 && !isdigit(*name))
-	*q++ = *name++;
-    *q = 0;
-    unit = atoi(name);
+    slprintf(ifdev, sizeof(ifdev), "/dev/%s", name);
+    for (q = ifdev + strlen(ifdev); --q >= ifdev; )
+	if (!isdigit(*q))
+	    break;
+    unit = atoi(q+1);
+    q[1] = 0;
 
     /*
      * Open the device and do a DLPI attach and phys_addr_req.
      */
     iffd = open(ifdev, O_RDWR);
     if (iffd < 0) {
-	syslog(LOG_ERR, "Can't open %s: %m", ifdev);
+	error("Can't open %s: %m", ifdev);
 	return 0;
     }
     if (dlpi_attach(iffd, unit) < 0
@@ -1372,6 +1441,7 @@ get_hw_addr(name, hwaddr)
 #endif
     hwaddr->sa_family = AF_UNSPEC;
     memcpy(hwaddr->sa_data, adrp, adrlen);
+#endif /* 0 */
 
     return 1;
 }
@@ -1434,7 +1504,7 @@ dlpi_get_reply(fd, reply, expected_prim, maxlen)
 
     if (buf.len < sizeof(ulong)) {
 	if (debug)
-	    syslog(LOG_DEBUG, "dlpi response short (len=%d)\n", buf.len);
+	    dbglog("dlpi response short (len=%d)\n", buf.len);
 	return -1;
     }
 
@@ -1443,11 +1513,11 @@ dlpi_get_reply(fd, reply, expected_prim, maxlen)
 
     if (debug) {
 	if (reply->dl_primitive == DL_ERROR_ACK) {
-	    syslog(LOG_DEBUG, "dlpi error %d (unix errno %d) for prim %x\n",
+	    dbglog("dlpi error %d (unix errno %d) for prim %x\n",
 		   reply->error_ack.dl_errno, reply->error_ack.dl_unix_errno,
 		   reply->error_ack.dl_error_primitive);
 	} else {
-	    syslog(LOG_DEBUG, "dlpi unexpected response prim %x\n",
+	    dbglog("dlpi unexpected response prim %x\n",
 		   reply->dl_primitive);
 	}
     }
@@ -1494,7 +1564,7 @@ GetMask(addr)
     if (ifc.ifc_buf == 0)
 	return mask;
     if (ioctl(ipfd, SIOCGIFCONF, &ifc) < 0) {
-	syslog(LOG_WARNING, "Couldn't get system interface list: %m");
+	warn("Couldn't get system interface list: %m");
 	free(ifc.ifc_buf);
 	return mask;
     }
@@ -1511,7 +1581,7 @@ GetMask(addr)
 	/*
 	 * Check that the interface is up, and not point-to-point or loopback.
 	 */
-	strncpy(ifreq.ifr_name, ifr->ifr_name, sizeof(ifreq.ifr_name));
+	strlcpy(ifreq.ifr_name, ifr->ifr_name, sizeof(ifreq.ifr_name));
 	if (ioctl(ipfd, SIOCGIFFLAGS, &ifreq) < 0)
 	    continue;
 	if ((ifreq.ifr_flags & (IFF_UP|IFF_POINTOPOINT|IFF_LOOPBACK))
@@ -1540,9 +1610,9 @@ logwtmp(line, name, host)
 
     if (name[0] != 0) {
 	/* logging in */
-	strncpy(utmpx.ut_user, name, sizeof(utmpx.ut_user));
-	strncpy(utmpx.ut_id, ifname, sizeof(utmpx.ut_id));
-	strncpy(utmpx.ut_line, line, sizeof(utmpx.ut_line));
+	strlcpy(utmpx.ut_user, name, sizeof(utmpx.ut_user));
+	strlcpy(utmpx.ut_id, ifname, sizeof(utmpx.ut_id));
+	strlcpy(utmpx.ut_line, line, sizeof(utmpx.ut_line));
 	utmpx.ut_pid = getpid();
 	utmpx.ut_type = USER_PROCESS;
     } else {
@@ -1561,7 +1631,7 @@ get_host_seed()
     char buf[32];
 
     if (sysinfo(SI_HW_SERIAL, buf, sizeof(buf)) < 0) {
-	syslog(LOG_ERR, "sysinfo: %m");
+	error("sysinfo: %m");
 	return 0;
     }
     return (int) strtoul(buf, NULL, 16);
@@ -1581,7 +1651,7 @@ strioctl(fd, cmd, ptr, ilen, olen)
     if (ioctl(fd, I_STR, &str) == -1)
 	return -1;
     if (str.ic_len != olen)
-	syslog(LOG_DEBUG, "strioctl: expected %d bytes, got %d for cmd %x\n",
+	dbglog("strioctl: expected %d bytes, got %d for cmd %x\n",
 	       olen, str.ic_len, cmd);
     return 0;
 }
@@ -1602,15 +1672,16 @@ lock(dev)
     char ascii_pid[12];
 
     if (stat(dev, &sbuf) < 0) {
-	syslog(LOG_ERR, "Can't get device number for %s: %m", dev);
+	error("Can't get device number for %s: %m", dev);
 	return -1;
     }
     if ((sbuf.st_mode & S_IFMT) != S_IFCHR) {
-	syslog(LOG_ERR, "Can't lock %s: not a character device", dev);
+	error("Can't lock %s: not a character device", dev);
 	return -1;
     }
-    sprintf(lock_file, "%s%03d.%03d.%03d", LOCK_PREFIX, major(sbuf.st_dev),
-	    major(sbuf.st_rdev), minor(sbuf.st_rdev));
+    slprintf(lock_file, sizeof(lock_file), "%s%03d.%03d.%03d",
+	     LOCK_PREFIX, major(sbuf.st_dev),
+	     major(sbuf.st_rdev), minor(sbuf.st_rdev));
 
     while ((fd = open(lock_file, O_EXCL | O_CREAT | O_RDWR, 0644)) < 0) {
 	if (errno == EEXIST
@@ -1618,7 +1689,7 @@ lock(dev)
 	    /* Read the lock file to find out who has the device locked */
 	    n = read(fd, ascii_pid, 11);
 	    if (n <= 0) {
-		syslog(LOG_ERR, "Can't read pid from lock file %s", lock_file);
+		error("Can't read pid from lock file %s", lock_file);
 		close(fd);
 	    } else {
 		ascii_pid[n] = 0;
@@ -1627,24 +1698,24 @@ lock(dev)
 		    /* pid no longer exists - remove the lock file */
 		    if (unlink(lock_file) == 0) {
 			close(fd);
-			syslog(LOG_NOTICE, "Removed stale lock on %s (pid %d)",
+			notice("Removed stale lock on %s (pid %d)",
 			       dev, pid);
 			continue;
 		    } else
-			syslog(LOG_WARNING, "Couldn't remove stale lock on %s",
+			warn("Couldn't remove stale lock on %s",
 			       dev);
 		} else
-		    syslog(LOG_NOTICE, "Device %s is locked by pid %d",
+		    notice("Device %s is locked by pid %d",
 			   dev, pid);
 	    }
 	    close(fd);
 	} else
-	    syslog(LOG_ERR, "Can't create lock file %s: %m", lock_file);
+	    error("Can't create lock file %s: %m", lock_file);
 	lock_file[0] = 0;
 	return -1;
     }
 
-    sprintf(ascii_pid, "%10d\n", getpid());
+    slprintf(ascii_pid, sizeof(ascii_pid), "%10d\n", getpid());
     write(fd, ascii_pid, 11);
 
     close(fd);
@@ -1682,9 +1753,167 @@ cifroute(u, our, his)
     rt.rt_flags = RTF_HOST;
 
     if (ioctl(ipfd, SIOCDELRT, &rt) < 0) {
-	syslog(LOG_ERR, "Can't delete route: %m");
+	error("Can't delete route: %m");
 	return 0;
     }
 
     return 1;
 }
+
+/*
+ * have_route_to - determine if the system has a route to the specified
+ * IP address.  Returns 0 if not, 1 if so, -1 if we can't tell.
+ * `addr' is in network byte order.
+ * For demand mode to work properly, we have to ignore routes
+ * through our own interface.
+ */
+#ifndef T_CURRENT		/* needed for Solaris 2.5 */
+#define T_CURRENT	MI_T_CURRENT
+#endif
+
+int
+have_route_to(addr)
+    u_int32_t addr;
+{
+    int fd, r, flags, i;
+    struct {
+	struct T_optmgmt_req req;
+	struct opthdr hdr;
+    } req;
+    union {
+	struct T_optmgmt_ack ack;
+	unsigned char space[64];
+    } ack;
+    struct opthdr *rh;
+    struct strbuf cbuf, dbuf;
+    int nroutes;
+    mib2_ipRouteEntry_t routes[8];
+    mib2_ipRouteEntry_t *rp;
+
+    fd = open("/dev/ip", O_RDWR);
+    if (fd < 0) {
+	warn("have_route_to: couldn't open /dev/ip: %m");
+	return -1;
+    }
+
+    req.req.PRIM_type = T_OPTMGMT_REQ;
+    req.req.OPT_offset = (char *) &req.hdr - (char *) &req;
+    req.req.OPT_length = sizeof(req.hdr);
+    req.req.MGMT_flags = T_CURRENT;
+
+    req.hdr.level = MIB2_IP;
+    req.hdr.name = 0;
+    req.hdr.len = 0;
+
+    cbuf.buf = (char *) &req;
+    cbuf.len = sizeof(req);
+
+    if (putmsg(fd, &cbuf, NULL, 0) == -1) {
+	warn("have_route_to: putmsg: %m");
+	close(fd);
+	return -1;
+    }
+
+    for (;;) {
+	cbuf.buf = (char *) &ack;
+	cbuf.maxlen = sizeof(ack);
+	dbuf.buf = (char *) routes;
+	dbuf.maxlen = sizeof(routes);
+	flags = 0;
+	r = getmsg(fd, &cbuf, &dbuf, &flags);
+	if (r == -1) {
+	    warn("have_route_to: getmsg: %m");
+	    close(fd);
+	    return -1;
+	}
+
+	if (cbuf.len < sizeof(struct T_optmgmt_ack)
+	    || ack.ack.PRIM_type != T_OPTMGMT_ACK
+	    || ack.ack.MGMT_flags != T_SUCCESS
+	    || ack.ack.OPT_length < sizeof(struct opthdr)) {
+	    dbglog("have_route_to: bad message len=%d prim=%d",
+		   cbuf.len, ack.ack.PRIM_type);
+	    close(fd);
+	    return -1;
+	}
+
+	rh = (struct opthdr *) ((char *)&ack + ack.ack.OPT_offset);
+	if (rh->level == 0 && rh->name == 0)
+	    break;
+	if (rh->level != MIB2_IP || rh->name != MIB2_IP_21) {
+	    while (r == MOREDATA)
+		r = getmsg(fd, NULL, &dbuf, &flags);
+	    continue;
+	}
+
+	for (;;) {
+	    nroutes = dbuf.len / sizeof(mib2_ipRouteEntry_t);
+	    for (rp = routes, i = 0; i < nroutes; ++i, ++rp) {
+		if (rp->ipRouteMask != ~0) {
+		    dbglog("have_route_to: dest=%x gw=%x mask=%x\n",
+			   rp->ipRouteDest, rp->ipRouteNextHop,
+			   rp->ipRouteMask);
+		    if (((addr ^ rp->ipRouteDest) & rp->ipRouteMask) == 0
+			&& rp->ipRouteNextHop != remote_addr)
+			return 1;
+		}
+	    }
+	    if (r == 0)
+		break;
+	    r = getmsg(fd, NULL, &dbuf, &flags);
+	}
+    }
+    close(fd);
+    return 0;
+}
+
+/*
+ * get_pty - get a pty master/slave pair and chown the slave side to
+ * the uid given.  Assumes slave_name points to MAXPATHLEN bytes of space.
+ */
+int
+get_pty(master_fdp, slave_fdp, slave_name, uid)
+    int *master_fdp;
+    int *slave_fdp;
+    char *slave_name;
+    int uid;
+{
+    int mfd, sfd;
+    char *pty_name;
+    struct termios tios;
+
+    mfd = open("/dev/ptmx", O_RDWR);
+    if (mfd < 0) {
+	error("Couldn't open pty master: %m");
+	return 0;
+    }
+
+    pty_name = ptsname(mfd);
+    if (pty_name == NULL) {
+	error("Couldn't get name of pty slave");
+	close(mfd);
+	return 0;
+    }
+    if (chown(pty_name, uid, -1) < 0)
+	warn("Couldn't change owner of pty slave: %m");
+    if (chmod(pty_name, S_IRUSR | S_IWUSR) < 0)
+	warn("Couldn't change permissions on pty slave: %m");
+    if (unlockpt(mfd) < 0)
+	warn("Couldn't unlock pty slave: %m");
+
+    sfd = open(pty_name, O_RDWR);
+    if (sfd < 0) {
+	error("Couldn't open pty slave %s: %m", pty_name);
+	close(mfd);
+	return 0;
+    }
+    if (ioctl(sfd, I_PUSH, "ptem") < 0)
+	warn("Couldn't push ptem module on pty slave: %m");
+
+    dbglog("Using %s", pty_name);
+    strlcpy(slave_name, pty_name, MAXPATHLEN);
+    *master_fdp = mfd;
+    *slave_fdp = sfd;
+
+    return 1;
+}
